/*
 *  GFileRunner         by Kyle Davis
 *  
 *  GFileRunner and all its pertaining source code is under the GNU General Public License.
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gdk/gdkkeysyms.h>

#include "callbacks.h"
#include "support.h"
#include "gfilerunner-sys.h"
#include "gfilerunner-cursor.h"
#include "gfilerunner-clipboard.h"
#include "gfilerunner-basic-icon.h"
#include "gfilerunner-vfs-object.h"
#include "gfilerunner-script.h"
#include "gfilerunner-template.h"
#include "gfilerunner-desktop.h"
#include "gfilerunner-file-properties-window.h"
#include "gfilerunner-theme-defs.h"

//TODO:  This in no way, shape, or form needs to go here...should be managed by the main loop..
extern gfilerunner_window* gfrwin;

pid_t preview_pid=NULL;
gboolean previewing;
int icon_xspacing=GFR_ICON_WIDTH+16;   //Icon horizontal spacing
int icon_yspacing=GFR_ICON_HEIGHT+32; //Icon vertical spacing

/*  TODO:  _ALL_ audio crap should be a module  */
gint start_prieview_ac3(const char* filename){   
  char* arg_list[]={"ac3dec", filename, NULL};
  return(spawnvp(arg_list));
}

gint start_prieview_mp3(const char* filename){
  char* arg_list[]={"mpg123", filename, NULL};
  return(spawnvp(arg_list));
}

gint start_prieview_wav(const char* filename){   
  char* arg_list[]={"sox", filename, "-t", "ossdsp", "/dev/dsp", NULL};
  return(spawnvp(arg_list));
}

//This will return if it had to kill anything or not.
gboolean stop_preview(){
  int retval=0;
  if(previewing==TRUE){
    if(preview_pid>0){
      kill(preview_pid,SIGTERM);
      waitpid(preview_pid,&retval,WUNTRACED);
      preview_pid=0;
    }
    previewing=FALSE;
    return TRUE;
  }
  return FALSE;
}

void preview_ac3(const char* filename){
  stop_preview();
  preview_pid=start_prieview_ac3(filename);
  if(preview_pid>0){
    previewing=TRUE;
  }else{
    previewing=FALSE;
  }
}

void preview_mp3(const char* filename){
  stop_preview();
  preview_pid=start_prieview_mp3(filename);
  if(preview_pid>0){
    previewing=TRUE;
  }else{
    previewing=FALSE;
  }
}

void preview_wav(const char* filename){
  stop_preview();
  preview_pid=start_prieview_wav(filename);
  if(preview_pid>0){
    previewing=TRUE;
  }else{
    previewing=FALSE;
  }
}

gint hack_canvas_events(GnomeCanvasItem *item, GdkEvent*event, gpointer data){
	static double x, y;
	double item_x;
	double item_y;
	if((!item)&&(!event)&&(!data)){
		x=0.0;
		y=0.0;
		return(FALSE);
	}
	item_x = event->button.x;
	item_y = event->button.y;
	gnome_canvas_window_to_world(GNOME_CANVAS(gfrwin->canvas),item_x,item_y,&item_x,&item_y);
	switch(event->type){
		case GDK_MOTION_NOTIFY:
			gnome_canvas_item_move(GNOME_CANVAS_ITEM(gfrwin->bghack),(double)(item_x-x), (double)(item_y-y));
			x=item_x;
			y=item_y;
			break;
		default:
			break;
	}
	return(FALSE);
}

gint bghack_events(GnomeCanvasItem *item, GdkEvent *event, gpointer data){
	double item_x, item_y;
	GdkCursor *fleur;
	GtkWidget*menu=NULL;
	GtkWidget*subitem=NULL;
	GtkWidget*submenu=NULL;
	GtkWidget*subitem2=NULL;
	static GdkEvent* last_event=NULL;
	static int dragging=FALSE;
	GList*listitem=NULL;
	gfilerunner_vfs_obj*vfsobj;
	char*location=gfrwin->clocation;
	item_x = event->button.x;
	item_y = event->button.y;
	gnome_canvas_item_w2i(GNOME_CANVAS_ITEM(gfrwin->canvas_group), &item_x, &item_y);
	if(!last_event)last_event=gdk_event_copy(event);
	switch(event->type){
		  case GDK_2BUTTON_PRESS:
				break;

		  case GDK_BUTTON_PRESS:
				switch(event->button.button){
					case 1:
						//TODO:  Don't flush if shift or ctrl are down
						gfilerunner_window_clear_selection(gfrwin);
						if(!gfrwin->lasso){
						       gfilerunner_window_show_lasso(gfrwin,item_x,item_y,item_x+1,item_y+1);
						       fleur=gdk_cursor_new(GDK_FLEUR);
						       gnome_canvas_item_grab(GNOME_CANVAS_ITEM(gfrwin->bghack),GDK_POINTER_MOTION_MASK|GDK_BUTTON_RELEASE_MASK,fleur,event->button.time);
						       gdk_cursor_destroy(fleur);
						}
						break;
					 case 3: //Right click
							gfilerunner_window_clear_selection(gfrwin);
						  menu=gtk_menu_new();
						{
							GList*files=NULL;
							GnomeVFSFileInfo*info=NULL;
							gfilerunner_template*gfrtemplate=NULL;
							char*script=NULL;
							char*ongoingerror=NULL;
							submenu=NULL;
							gnome_vfs_directory_list_load(&files,PACKAGE_TEMPLATES_DIR,GNOME_VFS_FILE_INFO_GET_MIME_TYPE|GNOME_VFS_FILE_INFO_FOLLOW_LINKS,NULL);
							while(files){
								info=(GnomeVFSFileInfo*)files->data;
								if((strcmp(info->name,".")!=0)&&(strcmp(info->name,"..")!=0)&&(strcmp(info->mime_type,"x-directory/normal")!=0)){
									script=g_concat_dir_and_file(PACKAGE_TEMPLATES_DIR,info->name);
									gfrtemplate=gfilerunner_template_new(script);
									if(!submenu){
										subitem=gtk_menu_item_new_with_label("New-");
										gtk_widget_ref(subitem);
										gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
										gtk_menu_append(GTK_MENU(menu),subitem);
										gtk_widget_show(subitem);
										submenu=gtk_menu_new();
										gtk_widget_ref(submenu);
										gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",submenu,(GtkDestroyNotify)gtk_widget_unref);
									}
									if(gfrtemplate){
										/* TODO: make it so when user clicks on errorenous template, show which dep failed along with all deps */
										ongoingerror=g_strjoinv(",",gfilerunner_template_check_dependancies(gfrtemplate));
										if(gfrtemplate->name){  /* [semi] make sure that we've got a real module made for GFileRunner */
											if(!strstr(ongoingerror,GFILERUNNER_TEMPLATE_ERROR)){
												GtkWidget*label=NULL;
												GtkWidget*pixmap=NULL;
												GdkPixmap*_pixmap=NULL;
												GdkBitmap*_mask=NULL;
												GdkPixbuf*mime_icon=NULL;
												subitem2=gtk_pixmap_menu_item_new();
												mime_icon=get_mime_icon(gfrtemplate->mime_type);
												if(!mime_icon)mime_icon=get_pixbuf(PACKAGE_THEMES_DIR "/" GFILERUNNER_DEFAULT_THEME "/" GFILERUNNER_DEFAULT_ICON);
												mime_icon=gdk_pixbuf_scale_simple(mime_icon,32,32,GDK_INTERP_HYPER);
												gdk_pixbuf_render_pixmap_and_mask(mime_icon,&_pixmap,&_mask,10);
												gdk_pixbuf_unref(mime_icon);
												pixmap=gtk_pixmap_new(_pixmap,_mask);
												gtk_widget_ref(pixmap);
												gtk_object_set_data_full(GTK_OBJECT(subitem2),"pixmap",pixmap,(GtkDestroyNotify)gtk_widget_unref);
												gtk_pixmap_menu_item_set_pixmap(GTK_PIXMAP_MENU_ITEM(subitem2),pixmap);
												gtk_widget_show(pixmap);
												label=gtk_accel_label_new(gfrtemplate->name);
												gtk_widget_ref(label);
												gtk_object_set_data_full(GTK_OBJECT(subitem2),"label",label,(GtkDestroyNotify)gtk_widget_unref);
												gtk_widget_show(label);
												gtk_container_add(GTK_CONTAINER(subitem2),label);
												gtk_widget_ref(subitem2);
												gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem2,(GtkDestroyNotify)gtk_widget_unref);
												gtk_signal_connect(GTK_OBJECT(subitem2),"activate",GTK_SIGNAL_FUNC(on_template_activate), (gpointer)gfilerunner_template_dup(gfrtemplate));
												gtk_menu_append(GTK_MENU(submenu),subitem2);
												gtk_widget_show(subitem2);
											}else{
												/* we have an error, so let the user know */
												char*errtxt=g_strconcat("ERROR: ",gfrtemplate->name,NULL);
												subitem2=gtk_menu_item_new_with_label(errtxt);
												gtk_widget_ref(subitem2);
												gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem2,(GtkDestroyNotify)gtk_widget_unref);
												//gtk_signal_connect(GTK_OBJECT(subitem2),"activate",GTK_SIGNAL_FUNC(on_script_activate), (gpointer)gfilerunner_script_dup(gfrtemplate));
												gtk_menu_append(GTK_MENU(submenu),subitem2);
												gtk_widget_show(subitem2);
												g_free(errtxt);
											}
										}
										g_free(ongoingerror);
										gfilerunner_template_free(gfrtemplate);
									}
									g_free(script);
								}
								files=files->next;
							}
							if(submenu){
								gtk_menu_item_set_submenu(GTK_MENU_ITEM(subitem),submenu);
							}
						}
					  subitem=gtk_menu_item_new_with_label("New Folder-");
					  gtk_menu_append(GTK_MENU(menu),subitem);
					  gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_newfolder_activate), data);
					  gtk_widget_show(subitem);
					{
						GList*files=NULL;
						GnomeVFSFileInfo*info=NULL;
						gfilerunner_script*gfrscript=NULL;
						char*script=NULL;
						int mime_type_offset=0;
						submenu=NULL;
						gnome_vfs_directory_list_load(&files,PACKAGE_SCRIPTS_DIR,GNOME_VFS_FILE_INFO_GET_MIME_TYPE|GNOME_VFS_FILE_INFO_FOLLOW_LINKS,NULL);
						while(files){
							info=(GnomeVFSFileInfo*)files->data;
							if((strcmp(info->name,".")!=0)&&(strcmp(info->name,"..")!=0)&&(strcmp(info->mime_type,"x-directory/normal")!=0)){
								script=g_concat_dir_and_file(PACKAGE_SCRIPTS_DIR,info->name);
								gfrscript=gfilerunner_script_new(script);
								if(!submenu){
									subitem=gtk_menu_item_new();
									gtk_widget_ref(subitem);
									gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
									gtk_menu_append(GTK_MENU(menu),subitem);
									gtk_widget_show(subitem);
									subitem=gtk_menu_item_new_with_label("Scripts-");
									gtk_widget_ref(subitem);
									gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
									gtk_menu_append(GTK_MENU(menu),subitem);
									gtk_widget_show(subitem);
									submenu=gtk_menu_new();
									gtk_widget_ref(submenu);
									gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",submenu,(GtkDestroyNotify)gtk_widget_unref);
								}
								if(gfrscript){
									if((gfrscript->name)&&(gfrscript->mime_types)){  /* [semi] make sure that we've got a real module made for GFileRunner */
										int selection_matches=0;
										int selection_total=0;
										int reverse_mimetype=0;
										//THIS IS BROKEN
										//TODO:  Get URI/C_TXT of location
										//TODO:  only local? or script should decide...fix code to work with non local filesystems....
										GnomeVFSFileInfo*tmpvfsinfo=gnome_vfs_file_info_new();
										gnome_vfs_get_file_info(location,tmpvfsinfo,GNOME_VFS_FILE_INFO_GET_MIME_TYPE);
										if(tmpvfsinfo){
											reverse_mimetype=0;
											mime_type_offset=0;
											while(gfrscript->mime_types[mime_type_offset]){
												if(reverse_mimetype>0){
													if(strcasecmp(gfrscript->mime_types[mime_type_offset],tmpvfsinfo->mime_type)==0){
														selection_matches--;
														break;
													}
													//break;
												}else{
													if(strcasecmp(gfrscript->mime_types[mime_type_offset],"internal/all")==0){
														reverse_mimetype++;
														selection_matches++;
														//break;
													}else{
														if(strcasecmp(gfrscript->mime_types[mime_type_offset],tmpvfsinfo->mime_type)==0){
															selection_matches++;
															break;
														}
													}
												}
												mime_type_offset++;
											}
											selection_total++;
										}
										listitem=listitem->next;
										if(selection_matches>=selection_total){
											subitem2=gtk_menu_item_new_with_label(gfrscript->name);
											gtk_widget_ref(subitem2);
											gtk_object_set_data_full(GTK_OBJECT(submenu), "dyamic_context_menu_item",subitem2,(GtkDestroyNotify)gtk_widget_unref);
											//gtk_signal_connect(GTK_OBJECT(GTK_LABEL(children->data)),"button_release_event",GTK_SIGNAL_FUNC(on_menuitem_button_press),(gpointer)gfilerunner_script_dup(gfrscript));
											//gtk_signal_connect(GTK_OBJECT(label),"button_press_event",GTK_SIGNAL_FUNC(on_menuitem_button_press),(gpointer)gfilerunner_script_dup(gfrscript));
											gtk_signal_connect(GTK_OBJECT(subitem2),"activate",GTK_SIGNAL_FUNC(on_script_activate), (gpointer)gfilerunner_script_dup(gfrscript));
											gtk_menu_append(GTK_MENU(submenu),subitem2);
											gtk_widget_show(subitem2);
										}
									}
									gfilerunner_script_free(gfrscript);
								}
								g_free(script);
							}
							files=files->next;
						}
						if(submenu){
							gtk_menu_item_set_submenu(GTK_MENU_ITEM(subitem),submenu);
						}
					}

						subitem=gtk_menu_item_new();
					  gtk_menu_append(GTK_MENU(menu),subitem);
					  gtk_widget_show(subitem);
					  subitem=gtk_menu_item_new_with_label("FIXME:  Arrange by->");
					  gtk_menu_append(GTK_MENU(menu),subitem);
					  gtk_widget_show(subitem);
					  subitem=gtk_menu_item_new();
					  gtk_menu_append(GTK_MENU(menu),subitem);
					  gtk_widget_show(subitem);
					  subitem=gtk_menu_item_new_with_label("Paste-");
					  gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_btnPaste_clicked), data);
					  gtk_menu_append(GTK_MENU(menu),subitem);
					  gtk_widget_show(subitem);
					  subitem=gtk_menu_item_new_with_label("FIXME: Paste Link");
					  gtk_menu_append(GTK_MENU(menu),subitem);
					  gtk_widget_show(subitem);
					  gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,event->button.button,event->button.time);
					  break;
				 default:
					  break;
				}
				if(gfrwin->icon_selection){
					 vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->icon_selection)-1);
					 basic_icon_unhighlight(vfsobj->icon);
					 gfrwin->icon_selection=NULL;
				}
				break;
	
		  case GDK_MOTION_NOTIFY:
				if(gfrwin->last_item){
					 vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->icon_selection)-1);
					 if(vfsobj) basic_icon_hide_preview(vfsobj->icon);
					 if((GPOINTER_TO_INT(gfrwin->last_item)-1)!=(GPOINTER_TO_INT(gfrwin->icon_selection)-1)){ 
						 vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->last_item)-1);
						 basic_icon_hide_preview(vfsobj->icon);
						 basic_icon_unhighlight(vfsobj->icon);
					 }
					 gfrwin->last_item=NULL;
				}
				stop_preview();
				if(gfrwin->lasso){
					gfilerunner_window_show_lasso(gfrwin,gfrwin->lasso->x1,gfrwin->lasso->y1,(gfloat)item_x,(gfloat)item_y);
					gfilerunner_window_clear_selection(gfrwin);
					gfilerunner_clipboard_set_prefix(gfrwin->selection,gfrwin->clocation);
					listitem=gfrwin->file_queue;
					while(listitem){
						vfsobj=(gfilerunner_vfs_obj*)listitem->data;
						if(((vfsobj->icon->x1+GFR_ICON_WIDTH2)>gfrwin->lasso->fx1)&&
						   ((vfsobj->icon->x1+GFR_ICON_WIDTH2)<gfrwin->lasso->fx2)&&
						   ((vfsobj->icon->y1+GFR_ICON_HEIGHT2)>gfrwin->lasso->fy1)&&
						   ((vfsobj->icon->y1+GFR_ICON_HEIGHT2)<gfrwin->lasso->fy2)){
							basic_icon_highlight(vfsobj->icon);
							gfilerunner_clipboard_prepend_file(gfrwin->selection,vfsobj,CLIPBOARD_COPY);
						}
						listitem=listitem->next;
					}
				}
				break;
      
		  case GDK_BUTTON_RELEASE:
			        if(gfrwin->lasso){
						gfilerunner_window_hide_lasso(gfrwin);
			        	gnome_canvas_item_ungrab(GNOME_CANVAS_ITEM(gfrwin->bghack), event->button.time);
			        	dragging=FALSE;
			        }
				break;
		  default:
				break;
	 }
	//These signals cause problems with my Click-And-Move detection
	if((event->type!=GDK_ENTER_NOTIFY)&&(event->type!=GDK_LEAVE_NOTIFY)){
		gdk_event_free(last_event);
		last_event=gdk_event_copy(event);
	}
	return FALSE;
}

//TODO:  Get right click on menu items working
gint on_menuitem_button_press(GtkWidget*widget,GdkEvent*event,gpointer data){
	fprintf(stderr,"event->button.button=%i\n",event->button.button);
	return(FALSE);
}

gint icon_events(GnomeCanvasItem *item, GdkEvent *event, gpointer data){
   static double x, y;
   double new_x, new_y;
   GdkCursor *fleur;
   static GdkEvent* last_event=NULL;
   static int dragging=FALSE;
   double item_x, item_y;
	gfilerunner_vfs_obj*vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(data));
	gfilerunner_vfs_obj*tmpvfsobj=NULL;
   GtkWidget *menu;
	GtkWidget *submenu;
   GtkWidget *subitem;
	GtkWidget *subitem2;
   gboolean retval=FALSE;
	GList*listitem=NULL;
   if((!vfsobj)||(!vfsobj->icon)||(!vfsobj->info)||(!vfsobj->info->name))return(TRUE);
   item_x = event->button.x;
   item_y = event->button.y;
   gnome_canvas_item_w2i(GNOME_CANVAS_ITEM(gfrwin->canvas_group), &item_x, &item_y);
   if(last_event==NULL){
	   last_event=gdk_event_copy(event);
   }
	switch(event->type){
		case GDK_KEY_PRESS:
			switch(event->key.keyval){
				case GDK_Return:
					on_open1_activate(NULL,data);
					break;
				case GDK_Page_Up:
					{
						GtkAdjustment*adj=gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(gfrwin->scrolledwindowRHS));
						gtk_adjustment_set_value(adj,adj->value-adj->page_size);
						gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(gfrwin->scrolledwindowRHS),adj);
					}
					break;
				case GDK_Page_Down:
					{
						GtkAdjustment*adj=gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(gfrwin->scrolledwindowRHS));
						gtk_adjustment_set_value(adj,adj->value+adj->page_size);
						gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(gfrwin->scrolledwindowRHS),adj);
					}
					break;
				default:
					fprintf(stderr,"FIXME:  UNHANDLED KEY:  KEY STATE:  %i\n",event->key.keyval);
					break;
			}
			break;
    case GDK_2BUTTON_PRESS:      x = item_x;
      y = item_y;
      switch(event->button.button){
        case 1:
			on_open1_activate(NULL,data);
			break;
      default:
         break;
      }
      retval=TRUE;
      break;

      case GDK_BUTTON_PRESS:
         x = item_x;
         y = item_y;
         if(!gfrwin->icon_selection)gfrwin->icon_selection=GINT_TO_POINTER(data)+1;
			if(!((event->key.keyval&GDK_Shift_L)||(event->key.keyval&GDK_Shift_R))){
				//TODO:  Finish this...it is horribly broken...
				tmpvfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->icon_selection-1));
				basic_icon_unhighlight(tmpvfsobj->icon);
			}
         gfrwin->icon_selection=GINT_TO_POINTER(data)+1;
			basic_icon_highlight(vfsobj->icon);
			if(!(vfsobj->info->type&GNOME_VFS_FILE_TYPE_DIRECTORY)){
				char*strbuf=g_strdup_printf("Size:  %s    MIME Type:  %s",gnome_vfs_format_file_size_for_display(vfsobj->info->size),vfsobj->info->mime_type);
				gnome_appbar_set_status(GNOME_APPBAR(gfrwin->appbar),strbuf);
				g_free(strbuf);
			}
			{
				gboolean shouldclear=TRUE;
				listitem=gfrwin->selection->data;
				while(listitem){
					tmpvfsobj=(gfilerunner_vfs_obj*)listitem->data;
					if(tmpvfsobj){
						if(strcmp(vfsobj->info->name,tmpvfsobj->info->name)==0) shouldclear=FALSE;
					}
					listitem=listitem->next;
				}
				if(shouldclear){
					gfilerunner_window_clear_selection(gfrwin);
					basic_icon_highlight(vfsobj->icon);
					gfilerunner_clipboard_set_prefix(gfrwin->selection,gfrwin->clocation);
					gfilerunner_clipboard_prepend_file(gfrwin->selection,(gpointer)vfsobj,CLIPBOARD_COPY);
				}
			}
         switch(event->button.button){
            case 2:  //Middle click
					break;
            case 1: //tmp left click
					//todo: if shift/ctrl click then don't flush.
					break;
            case 3: //Right click
				menu=gtk_menu_new();
				gtk_widget_ref(menu);
				subitem=gtk_menu_item_new_with_label("Open");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_open1_activate), data);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				//TODO:  IMAGES ONLY!  Should be a module...or built in module....modular form at least
				if(strncmp("image/",vfsobj->info->mime_type,6)==0){
					subitem=gtk_menu_item_new_with_label("FIXME: View");
					gtk_widget_ref(subitem);
					gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
					gtk_menu_append(GTK_MENU(menu),subitem);
					gtk_widget_show(subitem);
				}
				{
					GList*applist=gnome_vfs_mime_get_all_applications(vfsobj->info->mime_type);
					GnomeVFSMimeApplication*mimeapp=NULL;
					submenu=NULL;
					while(applist){
						mimeapp=(GnomeVFSMimeApplication*)applist->data;
						if(mimeapp){
							if(!submenu){
								subitem=gtk_menu_item_new();
								gtk_widget_ref(subitem);
								gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
								gtk_menu_append(GTK_MENU(menu),subitem);
								gtk_widget_show(subitem);
								subitem=gtk_menu_item_new_with_label("Open With-");
								gtk_widget_ref(subitem);
								gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
								gtk_menu_append(GTK_MENU(menu),subitem);
								gtk_widget_show(subitem);
								submenu=gtk_menu_new();
								gtk_widget_ref(submenu);
								gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",submenu,(GtkDestroyNotify)gtk_widget_unref);
							}
							subitem2=gtk_menu_item_new_with_label(mimeapp->name);
							gtk_widget_ref(subitem2);
							//TODO:  see FIXME in on_openwith_activate regarding the user_data to be passed.
							gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem2,(GtkDestroyNotify)gtk_widget_unref);
							gtk_signal_connect(GTK_OBJECT(subitem2),"activate",GTK_SIGNAL_FUNC(on_openwith_activate), (gpointer)mimeapp);
							gtk_menu_append(GTK_MENU(submenu),subitem2);
							gtk_widget_show(subitem2);
						}
						applist=applist->next;
					}
					if(submenu){
						gtk_menu_item_set_submenu(GTK_MENU_ITEM(subitem),submenu);
					}
				}
				{
					GList*files=NULL;
					GnomeVFSFileInfo*info=NULL;
					gfilerunner_script*gfrscript=NULL;
					char*script=NULL;
					int mime_type_offset=0;
					submenu=NULL;
					gnome_vfs_directory_list_load(&files,PACKAGE_SCRIPTS_DIR,GNOME_VFS_FILE_INFO_GET_MIME_TYPE|GNOME_VFS_FILE_INFO_FOLLOW_LINKS,NULL);
					while(files){
						info=(GnomeVFSFileInfo*)files->data;
						if((strcmp(info->name,".")!=0)&&(strcmp(info->name,"..")!=0)&&(strcmp(info->mime_type,"x-directory/normal")!=0)){
							script=g_concat_dir_and_file(PACKAGE_SCRIPTS_DIR,info->name);
							gfrscript=gfilerunner_script_new(script);
							if(!submenu){
								subitem=gtk_menu_item_new();
								gtk_widget_ref(subitem);
								gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
								gtk_menu_append(GTK_MENU(menu),subitem);
								gtk_widget_show(subitem);
								subitem=gtk_menu_item_new_with_label("Scripts-");
								gtk_widget_ref(subitem);
								gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
								gtk_menu_append(GTK_MENU(menu),subitem);
								gtk_widget_show(subitem);
								submenu=gtk_menu_new();
								gtk_widget_ref(submenu);
								gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",submenu,(GtkDestroyNotify)gtk_widget_unref);
							}
							if(gfrscript){
								if((gfrscript->name)&&(gfrscript->mime_types)){  /* [semi] make sure that we've got a real module made for GFileRunner */
									GList*listitem=gfrwin->selection->data;
									gfilerunner_vfs_obj*myvfsobj=NULL;
									int selection_matches=0;
									int selection_total=0;
									int reverse_mimetype=0;
									while(listitem){
										myvfsobj=listitem->data;
										if(myvfsobj){
											reverse_mimetype=0;
											mime_type_offset=0;
											while(gfrscript->mime_types[mime_type_offset]){
												if(reverse_mimetype>0){
													if(strcasecmp(gfrscript->mime_types[mime_type_offset],myvfsobj->info->mime_type)==0){
														selection_matches--;
														break;
													}
													//break;
												}else{
													if(strcasecmp(gfrscript->mime_types[mime_type_offset],"internal/all")==0){
														reverse_mimetype++;
														selection_matches++;
														//break;
													}else{
														if(strcasecmp(gfrscript->mime_types[mime_type_offset],myvfsobj->info->mime_type)==0){
															selection_matches++;
															break;
														}
													}
												}
												mime_type_offset++;
											}
											selection_total++;
										}
										listitem=listitem->next;
									}
									if(selection_matches>=selection_total){
										subitem2=gtk_menu_item_new_with_label(gfrscript->name);
										gtk_widget_ref(subitem2);
										gtk_object_set_data_full(GTK_OBJECT(submenu), "dyamic_context_menu_item",subitem2,(GtkDestroyNotify)gtk_widget_unref);
										//gtk_signal_connect(GTK_OBJECT(GTK_LABEL(children->data)),"button_release_event",GTK_SIGNAL_FUNC(on_menuitem_button_press),(gpointer)gfilerunner_script_dup(gfrscript));
										//gtk_signal_connect(GTK_OBJECT(label),"button_press_event",GTK_SIGNAL_FUNC(on_menuitem_button_press),(gpointer)gfilerunner_script_dup(gfrscript));
										gtk_signal_connect(GTK_OBJECT(subitem2),"activate",GTK_SIGNAL_FUNC(on_script_activate), (gpointer)gfilerunner_script_dup(gfrscript));
										gtk_menu_append(GTK_MENU(submenu),subitem2);
										gtk_widget_show(subitem2);
									}
								}
								gfilerunner_script_free(gfrscript);
							}
							g_free(script);
						}
						files=files->next;
					}
					if(submenu){
						gtk_menu_item_set_submenu(GTK_MENU_ITEM(subitem),submenu);
					}
				}
				subitem=gtk_menu_item_new();
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("Cut-");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_btnCut_clicked), data);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("FIXME: Copy");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_btnCopy_clicked), data);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("FIXME: Copy Location");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("Paste-");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_btnPaste_clicked), data);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("FIXME: Paste Link");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("Rename-");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_rename1_activate), data);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("FIXME: Duplicate");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("FIXME: Truncate");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("Move to Trash-");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_btnTrash_clicked), NULL);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
//TODO:  Another idea may be to "Edit symlink" on symlinks and "Fix symlink" on broken symlinks
/*				if(vfsobj->icon->decoration==GFR_ICON_DEC_SYMLINK){
						subitem=gtk_menu_item_new_with_label("FIXME: Follow Symlink");
						gtk_widget_ref(subitem);
						gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
						gtk_menu_append(GTK_MENU(menu),subitem);
						gtk_widget_show(subitem);
				}
*/
/*I know I had a good reason for wanting this....dunno what....but there was something.....*/
				subitem=gtk_menu_item_new();
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				subitem=gtk_menu_item_new_with_label("Properties-");
				gtk_widget_ref(subitem);
				gtk_object_set_data_full(GTK_OBJECT(menu), "dyamic_context_menu_item",subitem,(GtkDestroyNotify)gtk_widget_unref);
				gtk_signal_connect(GTK_OBJECT(subitem),"activate",GTK_SIGNAL_FUNC(on_properties1_activate), NULL);
				gtk_menu_append(GTK_MENU(menu),subitem);
				gtk_widget_show(subitem);
				gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,event->button.button,event->button.time);
				return(TRUE);
				break;
		default:
   	         break;
		}
		retval=TRUE;
		break;
	
      case GDK_MOTION_NOTIFY:
	//Left drag only for now
         if((!dragging)&&(last_event->type==GDK_BUTTON_PRESS)&&(last_event->button.button==1)){
            gnome_canvas_item_raise_to_top(item);
            fleur=gdk_cursor_new(GDK_FLEUR);
            gnome_canvas_item_grab(item,GDK_POINTER_MOTION_MASK|GDK_BUTTON_RELEASE_MASK,fleur,event->button.time);
            gdk_cursor_destroy(fleur);
            dragging=TRUE;
         }
//TODO:  Why so many variables in the first part of this if here???
         if(dragging){
/*				listitem=gfrwin->selection->data;
				while(listitem){
					tmpvfsobj=(gfilerunner_vfs_obj*)listitem->data;
					if(tmpvfsobj){
						new_x = item_x;
						new_y = item_y;
						//gnome_canvas_item_move(tmpvfsobj->icon->icon_group,(double)(new_x - x), (double)(new_y - y));
						basic_icon_move(tmpvfsobj->icon,(double)(new_x - x), (double)(new_y - y));
						x = new_x;
						y = new_y;
					}
					listitem=listitem->next;
				}*/
				new_x = item_x;
				new_y = item_y;
				basic_icon_move(vfsobj->icon,(double)(new_x - x), (double)(new_y - y));
				x = new_x;
				y = new_y;	
         }else{
            if(!gfrwin->last_item){
               basic_icon_prelight(vfsobj->icon);
            }else{
					if(gfrwin->last_item!=(GPOINTER_TO_INT(data)+1)){
						tmpvfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->last_item)-1);
						basic_icon_hide_preview(tmpvfsobj->icon);
//						if((GPOINTER_TO_INT(gfrwin->last_item)-1)!=(GPOINTER_TO_INT(gfrwin->icon_selection)-1)){
//							tmpvfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->last_item)-1);
							basic_icon_unhighlight(tmpvfsobj->icon);
						}
						gfilerunner_window_refresh_selection(gfrwin); //FIXME:  hack?
						basic_icon_prelight(vfsobj->icon);
						/* Auto-Preview Audio files here */
						//TODO:  Make this a toggle option....
						if(strcmp("audio/ac3",vfsobj->info->mime_type)==0){  //AC3
							char *fq_uri;
							fq_uri=g_concat_dir_and_file(gfrwin->clocation,vfsobj->info->name);
							basic_icon_show_preview(vfsobj->icon);
							preview_ac3(fq_uri);
							g_free(fq_uri);
						}
						if(strcmp("audio/x-mp3",vfsobj->info->mime_type)==0){   //MP3
							char *fq_uri;
							fq_uri=g_concat_dir_and_file(gfrwin->clocation,vfsobj->info->name);
							basic_icon_show_preview(vfsobj->icon);
							preview_mp3(fq_uri);
							g_free(fq_uri);
						}
						if(strcmp("audio/x-riff",vfsobj->info->mime_type)==0){  //WAV
							char *fq_uri;
							fq_uri=g_concat_dir_and_file(gfrwin->clocation,vfsobj->info->name);
							basic_icon_show_preview(vfsobj->icon);
							preview_wav(fq_uri);
							g_free(fq_uri);
						}
//FIXME:  TODO:  Get this to work correctly  (make it NOT restart when you move on the icon)
//					}
				}
         }
			gfrwin->last_item=GINT_TO_POINTER(data)+1;
         retval=TRUE;
         break;
      
      case GDK_BUTTON_RELEASE:
			if(dragging){
           gnome_canvas_item_raise_to_top(GNOME_CANVAS_ITEM(item));
           gnome_canvas_item_ungrab(GNOME_CANVAS_ITEM(item), event->button.time);
           dragging=FALSE;
         }
         retval=TRUE;
         break;

         default:
         break;
   }
   //These signals cause problems with my Click-And-Move detection
   if((event->type!=GDK_ENTER_NOTIFY)&&(event->type!=GDK_LEAVE_NOTIFY)){
      gdk_event_free(last_event);
      last_event=gdk_event_copy(event);
   }
   return(retval);
}

void set_basic_icon_signals(GnomeCanvasItem* item, gpointer data){
  gtk_signal_connect(GTK_OBJECT(item),"event",(GtkSignalFunc)icon_events, data);
}

gboolean on_gfilerunner_motion_notify_event(GtkWidget *widget,GdkEventMotion  *event, gpointer user_data){
  return FALSE;
}

void on_gfilerunner_check_resize(GtkWidget *widget,gpointer user_data){
  //printf("FIXME:  gfilerunner_resize event\n");
}

void on_new_file1_activate(GtkMenuItem     *menuitem,gpointer         user_data){
	fprintf(stderr,"broken?\n");
}

/*TODO: there should be a function to execute our script*/
void on_script_activate(GtkMenuItem*menuitem,gpointer user_data){
	gfilerunner_script*gfrscript=(gfilerunner_script*)user_data;
	GList*listitem=gfrwin->selection->data;
	char*location=gfrwin->clocation;
	gfilerunner_vfs_obj*vfsobj=NULL;
	GnomeVFSFileInfo*info=NULL;
	GnomeVFSURI* curi=NULL;
	GnomeVFSURI* puri=NULL;
	char*puri_txt=NULL;
	char*argstr=NULL;
	if(listitem){
		while(listitem){
			vfsobj=(gfilerunner_vfs_obj*)listitem->data;
			if(vfsobj){
				if(gfrscript->needs_terminal){
					//TODO:  make a parser function... ie get_cmd_safe_str that replaces:
					//	' ' = '\ '
					//	[ = \[
					//	] = \]
					//	( = \(
					//	) = \)
					argstr=g_strdup_printf("%s -e \"\\\"%s\\\" \\\"%s\\\" %i %i %i %i \\\"%s\\\" \\\"%s\\\" \\\"%s\\\"\" &",gfrprefs->locations->terminal,gfrscript->file,GFILERUNNER_SCRIPT_FLAG_EXECUTE,50,50,51,51,location,vfsobj->info->name,vfsobj->info->mime_type);
				}else{
					argstr=g_strdup_printf("\"%s\" \"%s\" %i %i %i %i \"%s\" \"%s\" \"%s\" &",gfrscript->file,GFILERUNNER_SCRIPT_FLAG_EXECUTE,50,50,51,51,location,vfsobj->info->name,vfsobj->info->mime_type);
				}
				fprintf(stderr,"FIXME:  running \"%s\"\n",argstr);
				system(argstr);
				g_free(argstr);
			}
			listitem=listitem->next;
		}
	}else{
		/*If there isn't anything selected we can assume that they right clicked on the background (the directory that they are in) for now...*/
		/*TODO:  This will need a gfrwin passed to it later as to determine which directory they are actually in....*/
		info=gnome_vfs_file_info_new();
		gnome_vfs_file_info_ref(info);
		gnome_vfs_get_file_info(location,info,GNOME_VFS_FILE_INFO_GET_MIME_TYPE);
		curi=gnome_vfs_uri_new(location);
		puri=gnome_vfs_uri_get_parent(curi);
		puri_txt=gnome_vfs_uri_to_string(puri,GNOME_VFS_URI_HIDE_USER_NAME&GNOME_VFS_URI_HIDE_PASSWORD&
																			GNOME_VFS_URI_HIDE_HOST_NAME&GNOME_VFS_URI_HIDE_HOST_PORT);
		gnome_vfs_uri_unref(puri);
		gnome_vfs_uri_unref(curi);
		if(gfrscript->local_only){ /* make sure that our uri string is local only */
			argstr=puri_txt; /* we'll use this as a temporary for now so we don't leak memory */
			puri_txt=gnome_vfs_get_local_path_from_uri(puri_txt);
			g_free(argstr);
		}
		if(gfrscript->needs_terminal){
			//TODO:  make a parser function... ie get_cmd_safe_str that replaces:
			//	' ' = '\ '
			//	[ = \[
			//	] = \]
			//	( = \(
			//	) = \)
			argstr=g_strdup_printf("%s -e \"\\\"%s\\\" \\\"%s\\\" %i %i %i %i \\\"%s\\\" \\\"%s\\\" \\\"%s\\\"\" &",gfrprefs->locations->terminal,gfrscript->file,GFILERUNNER_SCRIPT_FLAG_EXECUTE,50,50,51,51,puri_txt,info->name,info->mime_type);
		}else{
			argstr=g_strdup_printf("\"%s\" \"%s\" %i %i %i %i \"%s\" \"%s\" \"%s\" &",gfrscript->file,GFILERUNNER_SCRIPT_FLAG_EXECUTE,50,50,51,51,puri_txt,info->name,info->mime_type);
		}
		fprintf(stderr,"FIXME:  running \"%s\"\n",argstr);
		system(argstr);
		g_free(argstr);
		g_free(puri_txt);
		gnome_vfs_file_info_unref(info);
	}
	gfilerunner_script_free(gfrscript);
}

/*TODO: there should be a function to execute our script*/
void on_template_activate(GtkMenuItem*menuitem,gpointer user_data){
	gfilerunner_template*gfrtemplate=(gfilerunner_template*)user_data;
	char*location=gfrwin->clocation;
	char*argstr=NULL;
	char**deps=gfilerunner_template_check_dependancies(gfrtemplate);
	char*argdeplist=NULL;
	char*tmp=NULL;
	int depoffset=0;
	while(deps[depoffset]){
		fprintf(stderr,"%s\n",deps[depoffset]);
		if(!argdeplist){
			argdeplist=g_strdup_printf("\"%s\"",deps[depoffset]);
		}else{
			tmp=argdeplist;
			argdeplist=g_strdup_printf("%s \"%s\"",argdeplist,deps[depoffset]);
			g_free(tmp);
		}
		depoffset++;
	}  /* count how many deps there are */
	argstr=g_strdup_printf("\"%s\" \"%s\" \"%s\" %s &",gfrtemplate->file,GFILERUNNER_TEMPLATE_FLAG_EXECUTE,location,argdeplist);
fprintf(stderr,"FIXME:  running %s\n",argstr);
	system(argstr);
	g_free(argstr);
	g_free(argdeplist);
	gfilerunner_template_free(gfrtemplate);
}

void on_openwith_activate(GtkMenuItem* menuitem,gpointer user_data){
	GnomeVFSMimeApplication*mimeapp=(GnomeVFSMimeApplication*)user_data;
	GList*listitem=gfrwin->selection->data;
	gfilerunner_vfs_obj*vfsobj=NULL;
	char*location=gfrwin->clocation;
	char*argstr="";
	if(mimeapp->can_open_multiple_files){
		if(mimeapp->requires_terminal) argstr=g_strdup_printf("%s -x",gfrprefs->locations->terminal);
		argstr=g_strconcat(argstr,mimeapp->command,NULL);
		while(listitem){
			vfsobj=(gfilerunner_vfs_obj*)listitem->data;
			if(vfsobj){
				argstr=g_strconcat(argstr," \"",g_concat_dir_and_file(location,vfsobj->info->name),"\"",NULL);
			}
			listitem=listitem->next;
		}
		argstr=g_strconcat(argstr," &",NULL);
fprintf(stderr,"spawning:  %s\n",argstr);
		system(argstr);
		g_free(argstr);
	}else{
		while(listitem){
			vfsobj=(gfilerunner_vfs_obj*)listitem->data;
			if(vfsobj){
				if(mimeapp->requires_terminal){
					argstr=g_strdup_printf("%s -x %s \"%s\" &",gfrprefs->locations->terminal,mimeapp->command,g_concat_dir_and_file(location,vfsobj->info->name));
				}else{
					argstr=g_strdup_printf("%s \"%s\" &",mimeapp->command,g_concat_dir_and_file(location,vfsobj->info->name));
				}
fprintf(stderr,"spawning:  %s\n",argstr);
				system(argstr);
				g_free(argstr);
			}
			listitem=listitem->next;
		}
	}
}

void on_open1_activate(GtkMenuItem* menuitem,gpointer user_data){
	gfilerunner_vfs_obj*vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(user_data));
	GnomeVFSMimeApplication*vfsmimeapp=NULL;
	char*location=NULL;
	char*tmp=NULL;
	GList*listitem=NULL;
	if(!vfsobj){
		listitem=gfrwin->selection->data;
		while(listitem){
			if(listitem->data) vfsobj=(gfilerunner_vfs_obj*)listitem->data;
			//TODO:  Do it for every single one....for now, just the last one will do....
			listitem=listitem->next;
		}
	}
	if((!vfsobj)||(!vfsobj->icon)||(!vfsobj->info)||(!vfsobj->info->name))return;
	if(vfsobj->info->type==GNOME_VFS_FILE_TYPE_DIRECTORY){
		location=g_strdup(gfrwin->clocation);
		if(!ends_with_delim(location,GNOME_VFS_URI_PATH_CHR)){
			tmp=location;
			location=g_strconcat(tmp,GNOME_VFS_URI_PATH_STR,NULL);
			g_free(tmp);
		}
		tmp=location;
		location=g_strconcat(tmp,vfsobj->info->name,GNOME_VFS_URI_PATH_STR,NULL);
		g_free(tmp);
		gtk_entry_set_text(GTK_ENTRY(gfrwin->location),location);
		gfrwin->animation_handle=basic_icon_dup(vfsobj->icon);
		on_btnGo_clicked(NULL,NULL);
	}else{
		/*TODO:  check against ("text/"&& Exec Perms!) _not_ just "text/x" */
		if(strncmp(vfsobj->info->mime_type,"text/x",6)==0){
			char*msgstr=g_strdup_printf("File '%s' is a text file, but has executable permissions.\nWould you rather execute or open this file?",vfsobj->info->name);
			GtkWidget*question_dialog=gnome_message_box_new(msgstr,GNOME_MESSAGE_BOX_INFO,"Execute","Open as Text",NULL);
			int a=gnome_dialog_run(GNOME_DIALOG(question_dialog));
			g_free(msgstr);
			if(a==0){
				char*argstr=g_strdup_printf("%s -e \"%s\" &",gfrprefs->locations->terminal,g_concat_dir_and_file(gfrwin->clocation,vfsobj->info->name));
				system(argstr);
				g_free(argstr);
				return;
			}
		}
		vfsmimeapp=gnome_vfs_mime_get_default_application(vfsobj->info->mime_type);
		if(strcmp(vfsobj->info->mime_type,"application/x-executable-binary")==0){
			char*argstr=g_strdup_printf("%s &",g_concat_dir_and_file(gfrwin->clocation,vfsobj->info->name));
			system(argstr);
			g_free(argstr);
		}else{
			if(vfsmimeapp){
			  if(vfsmimeapp->command){
				char*argstr=g_strdup_printf("%s \"%s\" &",vfsmimeapp->command,g_concat_dir_and_file(gfrwin->clocation,vfsobj->info->name));
				system(argstr);
				g_free(argstr);
			  }
			}
		}
	}
}

void on_save1_activate(GtkMenuItem     *menuitem,gpointer         user_data){
}

void on_save_as1_activate(GtkMenuItem     *menuitem,gpointer         user_data){
}

void on_exit1_activate(GtkMenuItem *menuitem,gpointer user_data){
	stop_preview();
	on_btnStop_clicked(NULL,NULL);
	gfilerunner_preferences_sync(gfrprefs);
	gfilerunner_clipboard_free(gfrclipboard);
	free_common_pixbufs();
	g_list_free(gfrwin->file_queue); //TODO:  should free everything associated....also should be in gfilerunner_window_free
	gfilerunner_window_free(gfrwin); //TODO:  This should go into the global area...not here...
	gtk_exit(0);
}


void on_cut1_activate(GtkMenuItem *menuitem,gpointer user_data){
	on_btnCut_clicked(NULL,NULL);
}

void on_copy1_activate(GtkMenuItem *menuitem,gpointer user_data){
	on_btnCopy_clicked(NULL,NULL);
}

void on_paste1_activate(GtkMenuItem *menuitem,gpointer user_data){
	on_btnPaste_clicked(NULL,NULL);
}

void renamedlg_callback(char*text,gpointer data){
	if((text)&&(data)){
		char*location=gfrwin->clocation;
		gfilerunner_vfs_obj*vfsobj=(gfilerunner_vfs_obj*)data;
		//TODO:  Should be using a gfilerunner_vfs function...that way all moves are the same...
		//TODO:  error checking? or let the gfilerunner_vfs_move/rename function handle that?
		gnome_vfs_move(g_concat_dir_and_file(location,vfsobj->info->name),g_concat_dir_and_file(location,text),FALSE);
		basic_icon_set_caption(vfsobj->icon,text);
		g_free(vfsobj->info->name);
		vfsobj->info->name=g_strdup(text);
		gnome_vfs_set_file_info(g_concat_dir_and_file(location,text),vfsobj->info,GNOME_VFS_SET_FILE_INFO_NAME);
		fprintf(stderr,"FIXME:  TODO:  error checking in rename!  (replace w/ gfilerunner_vfs func)\n");
	}
}

void on_rename1_activate(GtkMenuItem*menuitem, gpointer user_data){
	gfilerunner_vfs_obj*vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->icon_selection)-1);
	GtkWidget*win=gnome_request_dialog(FALSE,"New Name:  ",vfsobj->info->name,255,
			(GnomeStringCallback)renamedlg_callback,(gpointer)vfsobj,(gpointer)gfrwin->window);
	gtk_widget_ref(win);
	gtk_object_set_data_full(GTK_OBJECT(win), "renamewin", win, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(win);
}

void on_delete1_activate(GtkMenuItem*menuitem,gpointer user_data){
	on_btnTrash_clicked(NULL,NULL);
}

void newfolderdlg_callback(gchar*string,gpointer data){
	if(string){
		char*location=gfrwin->clocation;
		char*newdir=g_concat_dir_and_file(location,string);
		fprintf(stderr,"FIXME:  made dir \"%s\" here!\n",newdir);
		//TODO:  gfilerunner_vfs_new_dir here instead (as to let our one func handle error checking....)
		gnome_vfs_make_directory(newdir,GNOME_VFS_PERM_USER_ALL|GNOME_VFS_PERM_GROUP_READ|GNOME_VFS_PERM_OTHER_READ);
	}
}

void on_newfolder_activate(GtkMenuItem*menuitem,gpointer user_data){
	GtkWidget*win=gnome_request_dialog(FALSE,"New Folder Name:  ","New Folder",255,
			(GnomeStringCallback)newfolderdlg_callback,(gpointer)gfrwin,GTK_WINDOW(gfrwin->window));
	gtk_widget_ref(win);
	gtk_object_set_data_full(GTK_OBJECT(win), "newfolderwin", win, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(win);
}

void on_properties1_activate(GtkMenuItem *menuitem,gpointer user_data){
	 gfilerunner_vfs_obj*vfsobj=NULL;
	 if(GPOINTER_TO_INT(gfrwin->icon_selection)){
		  vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(gfrwin->icon_selection)-1);
		  show_dlgFileProperties(vfsobj->info);
	 }
	 /*
	 if(selection){
	   show_properties_window((GList*)FileSelectedList);
	 }else{
	    //result=messagebox("you don't have anything selected\n");
		  //if result is nevershowagain; apply to settings (to never show this again)
	 }
	 */
}

void on_view1_antialiasing_activate(GtkMenuItem*menuitem, gpointer user_data){
//TODO:  FIND A BETTER WAY TO DO THIS!!!  ACTUALLY GET THE VALUE!!!
   gfrprefs->graphics->canvas_aa=gfrprefs->graphics->canvas_aa^TRUE;
}

void on_preferences1_activate(GtkMenuItem     *menuitem,gpointer         user_data){
}

void on_about1_activate(GtkMenuItem *menuitem,gpointer user_data){
	const char*authors[]={"Kyle Davis",NULL};
	GtkWidget*aboutbox=gnome_about_new("GFileRunner",VERSION,"Copyright(c) 2001-2002 Kyle Davis",authors,
																				"GFileRunner is GNOME-based file manager aimed to be a nicer, faster, and more powerful replacement for the current popular file manager, Nautilus.\n\n"
																				"Currently it does much of the things that Nautilus does, but faster. Some things are slower. Some things don't work. And some things aren't done.\n"
																				"If nothing else tell me why you hate it so that I can fix the problems.\nThanks for trying GFileRunner!",
																				PACKAGE_PIXMAPS_DIR "/gfilerunner.png");
	gtk_window_set_modal(GTK_WINDOW(aboutbox),TRUE);
	gtk_widget_show(aboutbox);
}

/*TODO:  This should be easier to change....like an options....or prefs menu....*/
gint animation_handler_callback(gpointer data){
   if((gfrwin->is_scanning==TRUE)&&(gfrwin->animation_handle)){
		basic_icon_switch_effect(gfrwin->animation_handle,gfrprefs->graphics->AniMode);
		switch(gfrprefs->graphics->AniMode){
			case ANI_BOUNCE:
				basic_icon_animate(gfrwin->animation_handle,4.0);
				break;
			case ANI_ROTATE:
				basic_icon_animate(gfrwin->animation_handle,10.0);
				break;
			case ANI_PULSATE:
				basic_icon_animate(gfrwin->animation_handle,0.05);
				break;
		}
   }
   return TRUE;
}

gint progress_handler_callback(gpointer user_data){
	char*strbuf=NULL;
	GtkProgress*progress=gnome_appbar_get_progress(GNOME_APPBAR(gfrwin->appbar));
	if((gfrwin->files_read>0)&&(gfrwin->total_objects>0)){
		gtk_progress_configure(GTK_PROGRESS(progress),gfrwin->files_read,0,gfrwin->total_objects);
	}
	strbuf=g_strdup_printf("Reading Files... (%i of %i)",gfrwin->files_read,gfrwin->total_objects);
	gnome_appbar_set_status(GNOME_APPBAR(gfrwin->appbar),strbuf);
	g_free(strbuf);
	return TRUE;
}

void set_background_color_ref(GtkWidget *canvas, GdkColor *color){
	/* try to alloc color */
	if(gdk_color_alloc(gdk_colormap_get_system(), color)){
		GtkStyle *style;
		style = gtk_style_copy(gtk_widget_get_style(GTK_WIDGET(canvas)));
		/* set new style */
		style->bg[GTK_STATE_NORMAL] = *color;
		gtk_widget_set_style(GTK_WIDGET(canvas), style);
	}
}

gint dequeue_handler_callback(gpointer user_data){
	char*location=NULL;
	char*uri=NULL;
	char*iconname=NULL;
	int xitems=(gfrwin->scrolledwindowRHS->allocation.width-20)/icon_xspacing;
	int x=(gfrwin->files_read%xitems)*icon_xspacing;
	int y=(((gfrwin->files_read-gfrwin->files_read%xitems)/xitems))*icon_yspacing;
	gfilerunner_vfs_obj*vfsobj=NULL;
	basic_icon_Decoration decoration=GFR_ICON_DEC_NONE;
	if((g_list_length(gfrwin->file_queue)>0)||(!gfrwin->is_scanning)){
   if(gfrwin->animation_idle_handle){
     gtk_timeout_remove(gfrwin->animation_idle_handle);
     gfrwin->animation_idle_handle=NULL;
     gfrwin->animation_handle=NULL;
	   if(gfrwin->icon_selection)gfrwin->icon_selection=NULL;
   // clear our canvas
     if(gfrwin->canvas_group){
       gtk_object_destroy(GTK_OBJECT(gfrwin->canvas_group));
       hack_canvas_events(NULL,NULL,NULL);
     }
     gtk_container_remove(GTK_CONTAINER(gfrwin->scrolledwindowRHS),gfrwin->canvas);
     gtk_object_destroy(GTK_OBJECT(gfrwin->canvas)); //destroy root canvas
     gdk_rgb_init();
     gtk_widget_push_visual(gdk_rgb_get_visual());
     gtk_widget_push_colormap(gdk_rgb_get_cmap());
     //Create a new root canvas depending on our AA option
	  //TODO:  Only do this when the user _CHANGES_ this option....
      if(gfrprefs->graphics->canvas_aa){
        gfrwin->canvas=gnome_canvas_new_aa();
      }else{
        gfrwin->canvas=gnome_canvas_new();
      }
      gtk_widget_pop_colormap();
      gtk_widget_pop_visual();
      gtk_widget_ref(gfrwin->canvas);
      gtk_object_set_data_full(GTK_OBJECT (gfrwin->window),"canvas",gfrwin->canvas,(GtkDestroyNotify)gtk_widget_unref);
      gnome_canvas_set_scroll_region(GNOME_CANVAS(gfrwin->canvas),0,0,gfrwin->canvas->allocation.width-20,((y>gfrwin->canvas->allocation.height)?y:gfrwin->canvas->allocation.height)+icon_yspacing);
      gnome_canvas_scroll_to(GNOME_CANVAS(gfrwin->canvas), 0, 0);
      gnome_canvas_set_pixels_per_unit(GNOME_CANVAS(gfrwin->canvas), 1.0);
      gtk_container_add(GTK_CONTAINER(gfrwin->scrolledwindowRHS),gfrwin->canvas);
      gnome_canvas_set_dither(GNOME_CANVAS(gfrwin->canvas),GDK_RGB_DITHER_NONE);
      gtk_widget_show(gfrwin->canvas);
//TODO:  User settable background.
//          OR  a solid color background (easiest and fastest)
//void set_background_color_ref(GtkWidget *canvas, GdkColor *color){
{
	GdkColor *color=g_new(GdkColor,1);
	color->pixel=0;
	color->red=63535;
	color->green=63535;
	color->blue=65535;
	set_background_color_ref(gfrwin->canvas,color);
	g_free(color);
}
      
   // make a new canvas group
     gfrwin->canvas_group=GNOME_CANVAS_GROUP(gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(gfrwin->canvas)),gnome_canvas_group_get_type(),
            "x",(gdouble)0.0, "y",(gdouble)0.0, NULL));		//This is nice....it will clear our canvas : )

	gfrwin->bghack=gnome_canvas_item_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),gnome_canvas_rect_get_type(),
				"x1",(gfloat)0, "y1",(gfloat)0,
				"x2",(gfloat)1, "y2",(gfloat)1,
				"fill_color_rgba",0xFFFFFF00,
				NULL);
	gnome_canvas_item_show(GNOME_CANVAS_ITEM(gfrwin->bghack));
	gtk_signal_connect(GTK_OBJECT(gfrwin->bghack),"event",(GtkSignalFunc)bghack_events,NULL);
	gtk_signal_connect(GTK_OBJECT(gfrwin->canvas),"event",(GtkSignalFunc)hack_canvas_events,NULL);
   }
    if(gfrwin->files_read<gfrwin->total_objects){
		vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,gfrwin->files_read);
      if(vfsobj->info->flags&GNOME_VFS_FILE_FLAGS_SYMLINK){
			GnomeVFSURI*vfsuri=NULL;
			decoration=GFR_ICON_DEC_SYMLINK;
			if(vfsobj->info->symlink_name[0]!=GNOME_VFS_URI_PATH_CHR){
				vfsuri=gnome_vfs_uri_new(gfrwin->clocation);
				vfsuri=gnome_vfs_uri_append_string(vfsuri,vfsobj->info->symlink_name);
			}else{
				vfsuri=gnome_vfs_uri_new(vfsobj->info->symlink_name);
			}
			if(gnome_vfs_uri_exists(vfsuri)){
				fprintf(stderr,"symlink:  %s -> %s\n",vfsobj->info->name,vfsobj->info->symlink_name);
			}else{
				fprintf(stderr,"broken-symlink:  %s is a broken symlink!\n",vfsobj->info->name);
				decoration=GFR_ICON_DEC_BROKEN_SYMLINK;
			}
			gnome_vfs_uri_unref(vfsuri);
		}
//TODO:  just make this a switch statement that changes a string for the desired icon.
      if((vfsobj->info->type&GNOME_VFS_FILE_TYPE_DIRECTORY)&&(strcmp(vfsobj->info->mime_type,"x-directory/normal")==0)){
			vfsobj->icon=basic_icon_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),vfsobj->info->name,
					get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_FOLDER),NULL,x,y,decoration);
      }else{
			if(vfsobj->info->mime_type==NULL)vfsobj->info->mime_type="internal/unknown";  //In case of a MIME detection failure.
			//If its an image file, attempt to use it as our icon.
			if((vfsobj->info->mime_type[0]=='i')&&(vfsobj->info->mime_type[1]=='m')&&(vfsobj->info->mime_type[2]=='a')&&(vfsobj->info->mime_type[3]=='g')&&(vfsobj->info->mime_type[4]=='e')){
				location=gfrwin->clocation;
				uri=g_malloc0((strlen(location)+strlen(vfsobj->info->name)+1));
				uri=strcpy(uri,location);
				if(!ends_with_delim(uri,GNOME_VFS_URI_PATH_CHR)){
					uri=strcat(uri,GNOME_VFS_URI_PATH_STR);
				}
				uri=strcat(uri,vfsobj->info->name);
				vfsobj->icon=basic_icon_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),vfsobj->info->name,uri,NULL,x,y,decoration);
				g_free(uri);
			}
			if(!vfsobj->icon){
				iconname=gnome_vfs_mime_get_icon(vfsobj->info->mime_type);
				if(iconname){
				iconname=gnome_vfs_icon_path_from_filename(iconname);
				if(iconname){
					 if(strcmp(iconname,gnome_unconditional_pixmap_file(""))!=0){
					cache_common_pixbuf(iconname);  //attempt to cache all MIME icons
					if((vfsobj->info->mime_type[0]=='a')&&(vfsobj->info->mime_type[1]=='u')&&(vfsobj->info->mime_type[2]=='d')&&(vfsobj->info->mime_type[3]=='i')&&(vfsobj->info->mime_type[4]=='o')){
						vfsobj->icon=basic_icon_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),vfsobj->info->name,iconname,
								get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_AUDIO_ICON_PREVIEW),x,y,decoration);
					}else{
						vfsobj->icon=basic_icon_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),vfsobj->info->name,iconname,NULL,x,y,decoration);
					}
					 }
				}
				}
			}
			//TODO:  INTERNAL MIME ICONS in case there aren't any MIME KEYS pointing to icons for some standard things
			if(!vfsobj->icon){      	   
				if((vfsobj->info->mime_type[0]=='a')&&(vfsobj->info->mime_type[1]=='u')&&(vfsobj->info->mime_type[2]=='d')&&(vfsobj->info->mime_type[3]=='i')&&(vfsobj->info->mime_type[4]=='o')){
					vfsobj->icon=basic_icon_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),vfsobj->info->name,
							get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_AUDIO_ICON),
							get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_AUDIO_ICON_PREVIEW),x,y,decoration);
				}
			}
			//TODO:  Handle Plain Text files
			//TODO:  Avoid using perms as fesible if the file is a symbolic link
			if(!vfsobj->icon){   //Check for executables
				if((vfsobj->info->permissions&GNOME_VFS_PERM_USER_EXEC)||(vfsobj->info->permissions&GNOME_VFS_PERM_OTHER_EXEC)){
					vfsobj->icon=basic_icon_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),vfsobj->info->name,
							get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_EXECUTABLE_ICON),NULL,x,y,decoration);
				}
			}
			if(!vfsobj->icon){      	   
				vfsobj->icon=basic_icon_new(GNOME_CANVAS_GROUP(gfrwin->canvas_group),vfsobj->info->name,
						get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_DEFAULT_ICON),NULL,x,y,decoration);
			}
      }
		  if(!vfsobj->icon){ //this should never happen
				printf("\nFIXME:  Sorry!...INSTALL GFileRunner into your GNOME prefix!!!\n");
				return TRUE;
		  }
		  set_basic_icon_signals(GNOME_CANVAS_ITEM(vfsobj->icon->icon_group),(gpointer)GINT_TO_POINTER(gfrwin->files_read));
		  gfrwin->files_read++;
		  gfrwin->objects_read++;  /* nuke files_read or make it actually work properly along w/ dirs_read */
		}else{
					 char*tmp;
					 if(gfrwin->dequeue_handle) gtk_idle_remove(gfrwin->dequeue_handle);
					 gfrwin->dequeue_handle=NULL;
					 if(gfrwin->progress_handle) gtk_idle_remove(gfrwin->progress_handle);
					 gfrwin->progress_handle=NULL;
					 printf("FIXME:  %i of %i objects were read in.\n",gfrwin->files_read,gfrwin->total_objects);
					 gnome_vfs_async_cancel(gfrwin->asyncvfs_handle);
					 set_cursor_normal(gfrwin->window);
					 gnome_appbar_clear_stack(GNOME_APPBAR(gfrwin->appbar));
					 tmp=g_strdup_printf("Done. (%i folders, %i files)",gfrwin->total_dirs,gfrwin->total_files);
					 gnome_appbar_set_status(GNOME_APPBAR(gfrwin->appbar),tmp);
					 gtk_widget_set_sensitive(gfrwin->btnStop,FALSE);
					 g_free(tmp);
		  }
	 }
  if((gfrwin->objects_read%8==0)||(gfrwin->objects_read>=gfrwin->total_objects)){
      gnome_canvas_set_scroll_region(GNOME_CANVAS(gfrwin->canvas),0,0,gfrwin->canvas->allocation.width-20,((y>gfrwin->canvas->allocation.height)?y:gfrwin->canvas->allocation.height)+icon_yspacing);
  }
  return TRUE;
}

void async_loading_callback(GnomeVFSAsyncHandle *handle,GnomeVFSResult result,GList *list,guint entries_read,gpointer callback_data){
	int a;
	GnomeVFSFileInfo *vfsFileInfo=NULL;
	GList*listitem;
	char* location=gfrwin->clocation;
	GnomeVFSURI* curi=NULL;
	char* urilocation=NULL;
	gfilerunner_vfs_obj*vfsobj=NULL;
	if((result!=GNOME_VFS_OK)&&(result!=GNOME_VFS_ERROR_EOF)){
		//TODO:  ftp://   FTP get UN/PW here!
		//TODO:  smb://   ** WARNING **: Couldn't find a master browser
		char*msgstr=g_strdup_printf("Unable to read from \"%s\"!\n Please try again or check that you have permission to access this object.",location);
		GtkWidget*question_dialog=gnome_message_box_new(msgstr,GNOME_MESSAGE_BOX_INFO,GNOME_STOCK_BUTTON_OK,NULL);
		a=gnome_dialog_run(GNOME_DIALOG(question_dialog));
	        g_free(msgstr);
//		on_btnStop_clicked(NULL,NULL);
		return;
	}
	if((gfrwin->total_files==0)&&(gfrwin->total_dirs==0))gfrwin->is_scanning=TRUE;
	gfrwin->total_objects+=entries_read;
	listitem=list;
	while(listitem){
		vfsFileInfo=(GnomeVFSFileInfo*)listitem->data;
//Add file to the queue
//TODO:  Sort via MIME_info, this will also cause it to be fresh in ram for the _next_ step....drawing!
//TODO:  The BETTER way to do this is to use  g_list_insert_sorted(GList* list,gpointer data,GCompareFunc func)
		vfsobj=gfilerunner_vfs_obj_new("FIXME: INCOMPLETE",vfsFileInfo,NULL);
		gnome_vfs_file_info_ref(vfsobj->info);
		if((vfsFileInfo->type&GNOME_VFS_FILE_TYPE_DIRECTORY)&&(strcmp(vfsFileInfo->mime_type,"x-directory/normal")==0)){
			gfrwin->file_queue=g_list_insert(gfrwin->file_queue,(gpointer)vfsobj,gfrwin->total_dirs);
			gfrwin->total_dirs++;
			//Add to tree
			curi=gnome_vfs_uri_new(location);        //Current URI
			curi=gnome_vfs_uri_append_string(curi,vfsFileInfo->name);
			urilocation=gnome_vfs_uri_to_string(curi,GNOME_VFS_URI_HIDE_USER_NAME&GNOME_VFS_URI_HIDE_PASSWORD&
					GNOME_VFS_URI_HIDE_HOST_NAME&GNOME_VFS_URI_HIDE_HOST_PORT);
			gfilerunner_tree_register_folder(gfrwin->gfrtree,urilocation);
			g_free(urilocation);
			gnome_vfs_uri_unref(curi);
		}else{
			gfrwin->file_queue=g_list_append(gfrwin->file_queue,(gpointer)vfsobj);
			gfrwin->total_files++;
		}
		listitem=listitem->next;
	}
	if(result==GNOME_VFS_ERROR_EOF){
		int priority=G_PRIORITY_DEFAULT_IDLE;
		/*TODO:  Auto-Turbo Mode threshold should be in options (recommend higher for slower computers, explain _what it does_)*/
		if((gfrwin->total_files+gfrwin->total_dirs)>=128) priority=G_PRIORITY_HIGH_IDLE+32;
		gfrwin->dequeue_handle=gtk_idle_add_priority(priority,dequeue_handler_callback,NULL);
		gfrwin->progress_handle=gtk_idle_add_priority(priority,progress_handler_callback,NULL);
		gfrwin->is_scanning=FALSE;
	}
}

void on_ctree_row_select(GtkCTree *ctree, GList *node, gint column, gpointer user_data){
	gfilerunner_tree_object*gfrtree_obj=(gfilerunner_tree_object*)gtk_ctree_node_get_row_data(GTK_CTREE(ctree),GTK_CTREE_NODE(node));
	if(strncmp(gfrtree_obj->location,"file://",7)==0){  //If its a file:// then strip the file:// from the head of the URI
		on_btnGo_clicked(NULL,(gpointer)(gfrtree_obj->location+7));
	}else{
		on_btnGo_clicked(NULL,(gpointer)gfrtree_obj->location);
	}
}

void on_btnGo_clicked(GtkButton*button,gpointer user_data){
   char*location=gtk_entry_get_text(GTK_ENTRY(gfrwin->location));
	GnomeVFSURI*curi=NULL;
   char*urilocation=NULL;
	GtkProgress*progress=gnome_appbar_get_progress(GNOME_APPBAR(gfrwin->appbar));
	GList*listitem=NULL;
	gfilerunner_vfs_obj*vfsobj=NULL;

	if(!user_data){
		location=gtk_entry_get_text(GTK_ENTRY(gfrwin->location));
	}else{
		location=(char*)user_data;
		gtk_entry_set_text(GTK_ENTRY(gfrwin->location),location);
	}
	curi=gnome_vfs_uri_new(location);
	urilocation=gnome_vfs_uri_to_string(curi,GNOME_VFS_URI_HIDE_USER_NAME&GNOME_VFS_URI_HIDE_PASSWORD&
						GNOME_VFS_URI_HIDE_HOST_NAME&GNOME_VFS_URI_HIDE_HOST_PORT);
	
/* stop anything else going on */
   on_btnStop_clicked(NULL,NULL);
	gtk_window_set_title(GTK_WINDOW(gfrwin->window),g_strdup_printf("GFileRunner - %s",location));
//TODO:  keep location as a URI and just display it using the proper humanize callback
	gfilerunner_tree_register_folder(gfrwin->gfrtree,urilocation);
	gnome_appbar_set_status(GNOME_APPBAR(gfrwin->appbar),"Loading...");
	gtk_progress_set_percentage(GTK_PROGRESS(progress),(gfloat)0.0);
	gnome_appbar_refresh(GNOME_APPBAR(gfrwin->appbar));
	set_cursor_busy(gfrwin->window);
	gfrwin->is_scanning=TRUE;  //we will be anyhow...
	gtk_widget_set_sensitive(gfrwin->btnStop,TRUE);
	gfrwin->animation_idle_handle=gtk_timeout_add(25,animation_handler_callback,NULL);
  
	gfrwin->files_read=0;
	gfrwin->dirs_read=0;
	gfrwin->objects_read=0;
	gfrwin->total_files=0;
	gfrwin->total_dirs=0;
	gfrwin->total_objects=0;
	gfrwin->last_item=NULL;

	gfrwin->clocation=g_strdup(location);

// Destroy old lists here....
  gfilerunner_window_clear_selection(gfrwin);
  if(gfrwin->file_queue){
	  listitem=gfrwin->file_queue;
	  while(listitem){
		  vfsobj=(gfilerunner_vfs_obj*)listitem->data;
		  if(vfsobj){
			  if(vfsobj->icon) basic_icon_free(vfsobj->icon);
			  if(vfsobj->info) gnome_vfs_file_info_unref(vfsobj->info);
			  gfilerunner_vfs_obj_free(vfsobj);
		  }
		  listitem=listitem->next;
	  }
	  g_list_free(gfrwin->file_queue);
  }
  gfrwin->file_queue=NULL;

  gnome_vfs_async_load_directory(&gfrwin->asyncvfs_handle,location,
			GNOME_VFS_FILE_INFO_DEFAULT|GNOME_VFS_FILE_INFO_GET_MIME_TYPE|GNOME_VFS_FILE_INFO_FOLLOW_LINKS,
			GNOME_VFS_DIRECTORY_FILTER_NONE,
			GNOME_VFS_DIRECTORY_FILTER_NOSELFDIR|GNOME_VFS_DIRECTORY_FILTER_NOPARENTDIR|GNOME_VFS_DIRECTORY_FILTER_NODOTFILES,
			"*",256,
			async_loading_callback,
			NULL);
  gnome_vfs_uri_unref(curi);
}

void on_btnBack_clicked(GtkButton *button,gpointer user_data){
}

void on_btnForward_clicked(GtkButton *button,gpointer user_data){
}

void on_btnUp_clicked(GtkButton *button,gpointer user_data){
  char* location=gfrwin->clocation;
  GnomeVFSURI* curi=gnome_vfs_uri_new(location);	//Current URI
  GnomeVFSURI* puri;					//Parent URI
  char* puri_txt=NULL;					//Parent URI - Text version
  if(gnome_vfs_uri_has_parent(curi)){
    puri=gnome_vfs_uri_get_parent(curi);
    puri_txt=gnome_vfs_get_local_path_from_uri(gnome_vfs_uri_to_string(puri,GNOME_VFS_URI_HIDE_NONE));
    if(puri_txt){
        if(!ends_with_delim(puri_txt,GNOME_VFS_URI_PATH_CHR)){
            puri_txt=strcat(puri_txt,GNOME_VFS_URI_PATH_STR);
        }
        on_btnGo_clicked(NULL,(gpointer)puri_txt);
	g_free(puri_txt);
    }
    gnome_vfs_uri_unref(puri);
    gnome_vfs_uri_unref(curi);
  }
}

void on_btnStop_clicked(GtkButton *button,gpointer user_data){
	if(gfrwin->is_scanning){
		gnome_vfs_async_cancel(gfrwin->asyncvfs_handle);

		if(gfrwin->animation_handle) basic_icon_reset_animation(gfrwin->animation_handle);

		gfrwin->is_scanning=FALSE;
	}
	set_cursor_normal(gfrwin->window);
	if(gfrwin->dequeue_handle) gtk_idle_remove(gfrwin->dequeue_handle);
	gfrwin->dequeue_handle=NULL;
	if(gfrwin->progress_handle) gtk_idle_remove(gfrwin->progress_handle);
	gfrwin->progress_handle=NULL;
	gfilerunner_window_hide_lasso(gfrwin);
	gtk_widget_set_sensitive(gfrwin->btnStop,FALSE);
	gnome_appbar_clear_stack(GNOME_APPBAR(gfrwin->appbar));
	gnome_appbar_set_status(GNOME_APPBAR(gfrwin->appbar),"Operation Cancelled by user.");
}

void on_btnRefresh_clicked(GtkButton *button,gpointer user_data){
	on_btnGo_clicked(NULL,NULL);
}

void on_btnCut_clicked(GtkButton *button,gpointer user_data){
	gfilerunner_vfs_obj*vfsobj=NULL;
	GList*item=NULL;
	if(gfilerunner_clipboard_has_data(gfrwin->selection)){
		gfilerunner_clipboard_flush(gfrclipboard);
		gfilerunner_clipboard_set_prefix(gfrclipboard,gfrwin->clocation);
		item=gfrwin->selection->data;
		while(item){
			vfsobj=(gfilerunner_vfs_obj*)item->data;
			if(vfsobj) gfilerunner_clipboard_prepend_file(gfrclipboard,(gpointer)gfilerunner_vfs_obj_dup(vfsobj),CLIPBOARD_CUT);
			item=item->next;
		}
	}
}

void on_btnCopy_clicked(GtkButton *button,gpointer user_data){
	gfilerunner_vfs_obj*vfsobj=NULL;
	GList*item=NULL;
	if(gfilerunner_clipboard_has_data(gfrwin->selection)){
		gfilerunner_clipboard_flush(gfrclipboard);
		gfilerunner_clipboard_set_prefix(gfrclipboard,gfrwin->clocation);
		item=gfrwin->selection->data;
		while(item){
			vfsobj=(gfilerunner_vfs_obj*)item->data;
			if(vfsobj) gfilerunner_clipboard_prepend_file(gfrclipboard,(gpointer)gfilerunner_vfs_obj_dup(vfsobj),CLIPBOARD_COPY);
			item=item->next;
		}
	}
}

void on_btnPaste_clicked(GtkButton *button,gpointer user_data){
	GList*listitem=gfrclipboard->data;
	gfilerunner_vfs_obj*vfsobj=NULL;
	char*location=gfrwin->clocation;
	char*dest_prefix=NULL;
	char*dest_uri=NULL;
	char*src_uri=NULL;
	GtkWidget *question_dialog;
	char*msgstr=NULL;
	gboolean should_refresh=FALSE;
	int ret=0;
	//TODO:  !!! Check to see if the clicked on a directory or not!!!
	//        (ie, right click, paste, and we copy/move our file into that directory!)
	//TODO:  Check list of special folders....give _module_defined_error_ if somebody is trying to move it....
	if(gfilerunner_clipboard_has_data(gfrclipboard)){
		if(user_data){	//User used the context menu to call this function....
			vfsobj=(gfilerunner_vfs_obj*)g_list_nth_data(gfrwin->file_queue,GPOINTER_TO_INT(user_data));
			if((vfsobj->info)&&(vfsobj->info->type==GNOME_VFS_FILE_TYPE_DIRECTORY)&&(strcmp(vfsobj->info->mime_type,"x-directory/normal")==0)){
				dest_prefix=g_concat_dir_and_file(location,vfsobj->info->name);
				if(!ends_with_delim(dest_prefix,GNOME_VFS_URI_PATH_CHR)) dest_prefix=g_strconcat(dest_prefix,GNOME_VFS_URI_PATH_STR,NULL);
			}
		}else{
			dest_prefix=location;
		}
		
fprintf(stderr,"clip_prefix=%s,dest_prefix=%s\n",gfrclipboard->uri_prefix,dest_prefix);
		if(strcmp(dest_prefix,location)==0) should_refresh=TRUE;
		while(listitem){
			vfsobj=(gfilerunner_vfs_obj*)listitem->data;
			dest_uri=g_concat_dir_and_file(dest_prefix,vfsobj->info->name);
			src_uri=g_concat_dir_and_file(gfrclipboard->uri_prefix,vfsobj->info->name);
			switch(gfrclipboard->mode){
				case CLIPBOARD_CUT:			//Move the files
					//TODO:  Use some sort of slick "file op in progress" interface here.
					printf("FIXME:  moving \'%s\'->\'%s\'!\n",src_uri,dest_uri);
					if(gnome_vfs_move(src_uri,dest_uri,FALSE)){
						msgstr=g_strdup_printf("Error moving \'%s\'!\nCheck to make sure you have access permissions and try again.",vfsobj->info->name);
						question_dialog=gnome_message_box_new(msgstr,GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
						ret=gnome_dialog_run(GNOME_DIALOG (question_dialog));
						g_free(msgstr);
					}
					if(strcmp(gfrclipboard->uri_prefix,location)==0){
						basic_icon_hide(vfsobj->icon);
					}
					break;
				case CLIPBOARD_COPY:		//Copy the files
					break;
//TODO:  If you end up using these you may want to make a gfrclipboard_handler(...) function or something
				case CLIPBOARD_DUPLICATE://Copy files with a file+"(Copy XXX)",duplicate_file_count
					break;
				case CLIPBOARD_TRASH:		//Move files to the trash bin
					break;
				case CLIPBOARD_DELETE:	//Delete files from the hard drive
					break;
			}
			//TODO:  gfilerunner_view should have a function that will allow append file...or something like that....
			if(should_refresh){
				//Update list size
				gfrwin->total_objects++;
				if((vfsobj->info->type&GNOME_VFS_FILE_TYPE_DIRECTORY)&&(strcmp(vfsobj->info->mime_type,"x-directory/normal")==0)){
					gfrwin->total_dirs++;
				}else{
					gfrwin->total_files++;
				}
				g_free(vfsobj->uri);
				vfsobj->uri=g_strdup(dest_prefix);
				//basic_icon_free(vfsobj->icon);
				//TODO:  Bad mem problem here....
				vfsobj->icon=NULL;
				gfrwin->file_queue=g_list_append(gfrwin->file_queue,gfilerunner_vfs_obj_dup(vfsobj));
			}	
			g_free(dest_uri);
			g_free(src_uri);
			listitem=listitem->next;
		}
		gfilerunner_clipboard_flush(gfrclipboard);
		if(should_refresh==TRUE){
		//We used to refresh...but thats costly....
//			on_btnGo_clicked(NULL,NULL);
		//Dequeue the new files in the list... 
			gfrwin->is_scanning=FALSE;
			gtk_idle_add_priority(G_PRIORITY_DEFAULT_IDLE,dequeue_handler_callback,NULL);
			gfrwin->progress_handle=gtk_idle_add_priority(G_PRIORITY_DEFAULT_IDLE,progress_handler_callback,NULL);
		}
	}else{
		printf("\nERROR:  clipboard is empty!\n");
	}
}

void on_btnTrash_clicked(GtkButton *button,gpointer user_data){
   //FIXME:  not done yet...
   //TODO:  make this run off of a seperate async move function that actually keeps the 
   //             progress bar updated as it moves the file....
   int ret=0;
   char* fq_uri;
   char* trash_uri;
   char* msgstr;
   GtkWidget *question_dialog;
	GList*listitem=gfrwin->selection->data;
	gfilerunner_vfs_obj*vfsobj=NULL;
	gfilerunner_vfs_obj*tmpvfsobj=NULL;
	char*location=gfrwin->clocation;
   if(user_data==NULL){
		while(listitem){
			vfsobj=(gfilerunner_vfs_obj*)listitem->data;
			if(vfsobj){
				on_btnTrash_clicked(NULL,(gpointer)vfsobj);
			}
			listitem=listitem->next;
		}
		return;
   }else{
		vfsobj=(gfilerunner_vfs_obj*)user_data;
	}
	if(vfsobj->info->name[0]==GNOME_VFS_URI_PATH_CHR){  //We have got an absolute filename
		fq_uri=g_strdup(vfsobj->info->name);
	}else{
		fq_uri=g_concat_dir_and_file(location,vfsobj->info->name);
	}

//																							info=g_new0(GnomeVFSFileInfo,1);
//																							gnome_vfs_get_file_info(fq_uri,info,GNOME_VFS_FILE_INFO_GET_MIME_TYPE);

	//TODO:  see if something in the trash is already called this....
   trash_uri=g_concat_dir_and_file(g_concat_dir_and_file(g_get_home_dir(),"Trash/"),vfsobj->info->name);
   printf("FIXME:  Moving '%s' to '%s'\n",fq_uri,trash_uri);
   if((vfsobj->info->flags&GNOME_VFS_FILE_FLAGS_SYMLINK)||(vfsobj->info->type==GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK)){
		//TODO:  Prompt for moving link to trash or moving actual file that link points to..(or both)
		msgstr=g_strdup_printf("Moving \"%s\"  -> \"%s\"\nThis file is a Symbolic Link.\nHow would you like to handle this file?",fq_uri,trash_uri);
      question_dialog=gnome_message_box_new(msgstr,GNOME_MESSAGE_BOX_QUESTION,"Just Trash Symbolic Link","Trash Symlink+Linked file",GNOME_STOCK_BUTTON_CANCEL,NULL);
      ret=gnome_dialog_run(GNOME_DIALOG(question_dialog));
      g_free(msgstr);
      switch(ret){
         case 0:  //Just Symlink
            if(gnome_vfs_move(fq_uri,trash_uri,FALSE)){
               question_dialog=gnome_message_box_new("Error moving file!\nCheck file permissions and try again.",GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
               ret=gnome_dialog_run(GNOME_DIALOG(question_dialog));
            }            
            break;
         case 1:  //Symlink+Linked file
            if(gnome_vfs_move(fq_uri,trash_uri,FALSE)){
               question_dialog=gnome_message_box_new("Error moving file!\nCheck file permissions and try again.",GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
               ret=gnome_dialog_run(GNOME_DIALOG (question_dialog));
            }
				g_free(fq_uri);
				if(vfsobj->info->symlink_name[0]==GNOME_VFS_URI_PATH_CHR){
					fq_uri=g_strdup(vfsobj->info->symlink_name);
				}else{
					fq_uri=g_concat_dir_and_file(location,vfsobj->info->symlink_name);
				}
				tmpvfsobj=gfilerunner_vfs_obj_new("FIXME: INCOMPLETE",gnome_vfs_file_info_new(),basic_icon_dup(vfsobj->icon));
				gnome_vfs_get_file_info(fq_uri,tmpvfsobj->info,GNOME_VFS_FILE_INFO_GET_MIME_TYPE);
            on_btnTrash_clicked(NULL,tmpvfsobj);
				//basic_icon_free(tmpvfsobj->icon);
				gfilerunner_vfs_obj_free(tmpvfsobj);
				//TODO:  We should update icons that are in the current folder...always
				//TODO:  Also we should check if a file exists before trying to remove it, ie symlinks that point to a file, when both are in the selection....u c......
				return;
            break;
         default: //Cancel
            ret=-1;
            break;
      }
   }else{
		if(gnome_vfs_move(fq_uri,trash_uri,FALSE)){
			question_dialog=gnome_message_box_new("Error moving file!\nCheck file permissions and try again.",GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
			ret=gnome_dialog_run(GNOME_DIALOG(question_dialog));
			ret=-1;
		}
   }
	g_free(fq_uri); //Dunno if this is correct.
   g_free(trash_uri);
   if(ret!=-1)basic_icon_hide(vfsobj->icon);
}

void on_btnUndo_clicked(GtkButton *button,gpointer user_data){
}

void on_btnSearch_clicked(GtkButton*button, gpointer user_data){
}

void on_btnHome_clicked(GtkButton *button,gpointer user_data){
  gtk_entry_set_text(GTK_ENTRY(gfrwin->location),g_get_home_dir());
  if(!ends_with_delim(gtk_entry_get_text(GTK_ENTRY(gfrwin->location)),GNOME_VFS_URI_PATH_CHR)){
    gtk_entry_append_text(GTK_ENTRY(gfrwin->location),GNOME_VFS_URI_PATH_STR);
  }
  on_btnGo_clicked(NULL,NULL);
}

void on_canvas_realize(GtkWidget *widget,GdkEvent *event,gpointer user_data){
   static gboolean firsttime=TRUE;
   if(firsttime){
		firsttime=FALSE;
      on_btnGo_clicked(NULL,NULL);
   }
}

gboolean on_canvas_destroy_event(GtkWidget *widget,GdkEvent *event,gpointer user_data){
  return FALSE;
}
