/*
 *  GFileRunner         by Kyle Davis
 *  
 *  GFileRunner and all its pertaining source code is under the GNU General Public License.
 *
 */

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf/gnome-canvas-pixbuf.h>
#include <libgnomeui/gnome-canvas.h>
#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <config.h>

#include "gfilerunner-basic-icon.h"
#include "gfilerunner-sys.h"
#include "gfilerunner-theme-defs.h"
#include "interface.h" /*this probably doesn't go here, but is needed for now for a global or two*/

GList* list_of_common_pixbufs=NULL;

GdkPoint* get_aspect_dimensions(int w, int h, int scaledwidth,int scaledheight){
   GdkPoint* tmp=(GdkPoint*)g_new0(GdkPoint,1);
   if(w>h){
     tmp->x=scaledwidth;
     tmp->y=(h*scaledheight)/w;
   }else{
     tmp->x=(w*scaledwidth)/h;
     tmp->y=scaledheight;
   }
   return(tmp);
}

//TODO:  Make this work for network files (at the users descretion) 
gboolean cache_common_pixbuf(const char*file){   //Return if it was cached or not
  GList* list=list_of_common_pixbufs;
  int a=0;
  if(file==NULL)return(FALSE);
  //make sure that its not already cached...
  while(list){
    if(strcmp(file,(char*)list->data)==0){
       //if so leave now
       return(FALSE);
    }
    a++;
    list=list->next;
  }
  fprintf(stderr,"Caching %s...",file);
  //if not cache it. and add to the list
  list_of_common_pixbufs=g_list_append(list_of_common_pixbufs,g_strdup(file));
  common_pixbufs=g_list_append(common_pixbufs,(gpointer)_get_pixbuf(file));
  //TODO:  Check for errors
  fprintf(stderr,"Done!\n");
  return(TRUE);
}

//Internal...absolutely will pull from file. (don't use this, see get_pixbuf)
GdkPixbuf* _get_pixbuf(const char*file){
	int w,h;
	int width=GFR_ICON_WIDTH;
	int height=GFR_ICON_HEIGHT;
	GdkPixbuf* pixbuf=NULL;
	GdkPixbuf* scaled_pixbuf=NULL;
	GdkPoint*scaled_size=NULL;
//	if((file==NULL)||(strcmp(file,"")==0))return(NULL);
	pixbuf=gdk_pixbuf_new_from_file(file);
	if(!pixbuf){
		char* filepath=gnome_vfs_icon_path_from_filename(file);
		if(filepath){
			pixbuf=gdk_pixbuf_new_from_file(filepath);
			g_free(filepath);
		}
	}
	if(pixbuf==NULL){
		FILE *f;
		f=fopen(file, "rb");
		if(f){
			pixbuf=rsvg_render_file(f, 0.6);  //0.6 = XICON_WIDTH / 48 //Nautilus's SVG scaling scheme
			fclose(f);
		}
	}
	//FIXME:  Scale down...
	if(pixbuf){
		w=gdk_pixbuf_get_width(pixbuf);
		h=gdk_pixbuf_get_height(pixbuf);
		if(w<=0)w=1;
		if(h<=0)h=1;
		//TODO:  Make this a toggle option  for possible speed improvement
		//(possibly not shown if too big....or only part shown....or something.....
		//definately size MIN & MAX border on image files though....
		if((w>width)||(h>height)){
			scaled_size=get_aspect_dimensions(w,h,width,height);
			if(pixbuf) scaled_pixbuf=gdk_pixbuf_scale_simple(pixbuf,scaled_size->x,scaled_size->y,GDK_INTERP_NEAREST);
			gdk_pixbuf_unref(pixbuf);
			pixbuf=scaled_pixbuf;
			g_free(scaled_size);
		}
	}
//TODO:  This should be an option for speed's sake
/*typedef enum {
  GDK_INTERP_NEAREST,
  GDK_INTERP_TILES,
  GDK_INTERP_BILINEAR,
  GDK_INTERP_HYPER
 } GdkInterpType;
*/
	return(pixbuf);
}

GdkPixbuf* get_pixbuf(const char*file){
  int a=0;
  //int w,h;
  //int width=64;
  //int height=64;
  GdkPixbuf* pixbuf=NULL;
  //GdkPixbuf* scaled_pixbuf=NULL;
  //GdkPoint*scaled_size=NULL;
  GList* list=list_of_common_pixbufs;
  if(!file)return(NULL);
  while(list){
    if(strcmp(file,(char*)list->data)==0){
    pixbuf=(GdkPixbuf*)g_list_nth_data(common_pixbufs,a);
    if(!pixbuf){
         return(NULL);
      }else{
         //TODO:  Maybe not copy this....maybe just link to the actuall data in memory....may require a basic_pixbuf structure
         return((GdkPixbuf*)gdk_pixbuf_copy(pixbuf));
      }
    }
    list=list->next;
    a++;
  }
  pixbuf=_get_pixbuf(file);
  return(pixbuf);
}

GdkPixbuf*get_mime_icon(const char*mime_type){
	GdkPixbuf*icon=NULL;
	char*iconname=NULL;
	iconname=gnome_vfs_mime_get_icon(mime_type);
	if(iconname){
		iconname=gnome_vfs_icon_path_from_filename(iconname);
		if(iconname){
			if(strcmp(iconname,gnome_unconditional_pixmap_file(""))!=0){
				cache_common_pixbuf(iconname);  /* attempt to cache all MIME icons */
				icon=get_pixbuf(iconname);
			}
		}
	}
	return(icon);
}

void free_common_pixbufs(){
  while(common_pixbufs){
    gdk_pixbuf_unref((GdkPixbuf*)common_pixbufs->data);
    common_pixbufs=common_pixbufs->next;
  }
  g_list_free(common_pixbufs);
  g_list_free(list_of_common_pixbufs);
  common_pixbufs=NULL;
  list_of_common_pixbufs=NULL;
}

basic_icon* basic_icon_new(const GnomeCanvasGroup*owner,const char*caption,const char*icon_filename,const char*preview_icon_filename,gdouble myX,gdouble myY,basic_icon_Decoration decoration){
/* Stipple for every other of pixel....this way it looks semi-transparent in GDK mode */
/* 0101 0101 = 170 */
  char mystipple[]={170,NULL};
  GdkBitmap*stipple=gdk_bitmap_create_from_data(NULL,mystipple,4,2);
  int w=0,h=0;
/*TODO:  make this a run time option */
	int width=GFR_ICON_WIDTH;
	int height=GFR_ICON_HEIGHT;
   int w2=GFR_ICON_WIDTH2;
   int h2=GFR_ICON_HEIGHT2;
  GdkPoint*scaled_size=NULL;
	basic_icon *icon=(basic_icon*)g_new0(basic_icon,1);
	icon->parent=GNOME_CANVAS_GROUP(owner);
	if(!icon->parent)return(NULL);

/* this causes grouping of the following objects */
	icon->icon_group=gnome_canvas_item_new(GNOME_CANVAS_GROUP(icon->parent),gnome_canvas_group_get_type(),"x",(gdouble)(myX+w2), "y",(gdouble)(myY+h2),NULL);
	if(!icon->icon_group)return(NULL);


	scaled_size=g_new0(GdkPoint,1);

/* icon */
	if(icon_filename){
		basic_icon_set_icon(icon,get_pixbuf(icon_filename));
	}else{
		basic_icon_set_icon(icon,get_pixbuf(GFILERUNNER_DEFAULT_ICON));
	}
	if(!icon->_pixbuf)return(NULL);

	//TODO:  Multiple decorations should be handled here....automagically that is.....
	icon->decoration=decoration;
	switch(decoration){
		case GFR_ICON_DEC_SYMLINK:
			icon->_link=get_pixbuf(get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_DECORATION_SYMLINK));
			icon->link=gnome_canvas_item_new(GNOME_CANVAS_GROUP(icon->icon_group),gnome_canvas_pixbuf_get_type(),
				"pixbuf",icon->_link,
				//Center the image if it is smaller than  width x height
				"x",(gdouble)width-w2-(w2>>1), "y",(gdouble)height-h2-(h2>>1),
				"width",(gdouble)width-w2-(w2>>1), "height",(gdouble)height-h2-(h2>>1),
				"width_set",TRUE, "height_set",TRUE,
				"x_in_pixels",TRUE,"y_in_pixels",TRUE,
				NULL);
			break;
		case GFR_ICON_DEC_BROKEN_SYMLINK:
			icon->_link=get_pixbuf(get_theme_ent(gfrprefs->theme->theme_prefix,GFILERUNNER_DECORATION_BROKEN_SYMLINK));
			icon->link=gnome_canvas_item_new(GNOME_CANVAS_GROUP(icon->icon_group),gnome_canvas_pixbuf_get_type(),
				"pixbuf",icon->_link,
				//Center the image if it is smaller than  width x height
				"x",(gdouble)width-w2-(w2>>1), "y",(gdouble)height-h2-(h2>>1),
				"width",(gdouble)width-w2-(w2>>1), "height",(gdouble)height-h2-(h2>>1),
				"width_set",TRUE, "height_set",TRUE,
				"x_in_pixels",TRUE,"y_in_pixels",TRUE,
				NULL);
			break;
		case GFR_ICON_DEC_NONE:
		default:
	}
	
	basic_icon_set_caption(icon,caption);
      
	//Draw a circle on our canvas group (used for the highlight)
	icon->highlight=gnome_canvas_item_new(GNOME_CANVAS_GROUP(icon->icon_group),gnome_canvas_ellipse_get_type(),
		"x1",(gfloat)-w2, "y1",(gfloat)-h2,
		"x2",(gfloat)w2, "y2",(gfloat)h2,
		"fill_color_rgba",0x2222FF22,
      "fill_stipple",stipple,
		"outline_color_rgba",0x00006622,
		"width_pixels",0.0,
		NULL);

//icon->preview
	//TODO:  Keep aspect ratio of pictures being used as icons
	if(preview_icon_filename){
   	icon->_preview_pixbuf=get_pixbuf(preview_icon_filename);
   	if(icon->_preview_pixbuf){
         w=gdk_pixbuf_get_width(icon->_preview_pixbuf);
         h=gdk_pixbuf_get_height(icon->_preview_pixbuf);
         if(w==0)w++;
         if(h==0)h++;
         scaled_size=get_aspect_dimensions(w,h,width,height);	   
      	icon->preview=gnome_canvas_item_new(GNOME_CANVAS_GROUP(icon->icon_group),gnome_canvas_pixbuf_get_type(),
      		"pixbuf",icon->_preview_pixbuf,
      	   //Center the image if it is smaller than  width x height
      		"x",(gdouble) ((scaled_size->x<width)?(width-scaled_size->x)>>1:0)-w2, "y",(gdouble)((scaled_size->y<height)?(height-scaled_size->y)>>1:0)-h2,
      		"width",(gdouble)scaled_size->x, "height",(gdouble)scaled_size->y,
      		"width_set",TRUE, "height_set",TRUE,
      		"x_in_pixels",TRUE,"y_in_pixels",TRUE,
      		NULL);
      	//Hide our preview by default
      	basic_icon_hide_preview(icon);
      }else{
         icon->preview=NULL;
      }
	}
		
	//Hide our highlight by default
	basic_icon_unprelight(icon);
   icon->x1=myX;
   icon->y1=myY;
   icon->x2=myX+width;
   icon->y2=myY+height;
   icon->angle=0.0;
   icon->scale=1.0;
   icon->visible=TRUE;
   icon->effect_type=ANI_ROTATE; //These are forced to be different here so that the second is the real one that will get the initialized vars
   basic_icon_switch_effect(icon,ANI_BOUNCE);
	g_free(scaled_size);
	return(icon);
}
void basic_icon_set_icon(basic_icon*icon,GdkPixbuf*pixbuf){
	GdkPoint*scaled_size=NULL;
	GdkPixbuf*scaled_pixbuf=NULL;
	int width=GFR_ICON_WIDTH;
	int height=GFR_ICON_HEIGHT;
   int w2=GFR_ICON_WIDTH2;
   int h2=GFR_ICON_HEIGHT2;
	int w,h;
	if((icon)&&(pixbuf)){
		scaled_size=g_new0(GdkPoint,1);
		w=gdk_pixbuf_get_width(pixbuf);
		h=gdk_pixbuf_get_height(pixbuf);
		if((w>width)||(h>height)){
			scaled_size=get_aspect_dimensions(w,h,width,height);
			if(pixbuf) scaled_pixbuf=gdk_pixbuf_scale_simple(pixbuf,scaled_size->x,scaled_size->y,GDK_INTERP_NEAREST);
		}else{
			scaled_pixbuf=gdk_pixbuf_copy(pixbuf);
			scaled_size->x=w;
			scaled_size->y=h;
		}
		if(scaled_pixbuf){
			if(icon->_pixbuf){
				gdk_pixbuf_unref(icon->_pixbuf);
				icon->_pixbuf=scaled_pixbuf;
				gnome_canvas_item_set(GNOME_CANVAS_ITEM(icon->pixbuf),
						"pixbuf",icon->_pixbuf,
						"x",(gdouble) ((scaled_size->x<width)?(width-scaled_size->x)>>1:0)-w2, "y",(gdouble)((scaled_size->y<height)?(height-scaled_size->y)>>1:0)-h2,
						NULL);
			}else{
				icon->_pixbuf=scaled_pixbuf;
				icon->pixbuf=gnome_canvas_item_new(GNOME_CANVAS_GROUP(icon->icon_group),gnome_canvas_pixbuf_get_type(),
				"pixbuf",icon->_pixbuf,
				//Center the image if it is smaller than  width x height
				"x",(gdouble) ((scaled_size->x<width)?(width-scaled_size->x)>>1:0)-w2, "y",(gdouble)((scaled_size->y<height)?(height-scaled_size->y)>>1:0)-h2,
//				"width",(gdouble)scaled_size->x, "height",(gdouble)scaled_size->y,
//				"width_set",TRUE, "height_set",TRUE,
				"width",(gdouble)width, "height",(gdouble)height,
				"width_set",FALSE, "height_set",FALSE,
				"x_in_pixels",TRUE,"y_in_pixels",TRUE,
				NULL);
			}
		}
		g_free(scaled_size);
	}
}
basic_icon* basic_icon_set_caption(basic_icon*icon,const char*caption){
  int a=0,w=0;
  char* tmp;
//TODO:  make this a run time option
	int width=GFR_ICON_WIDTH;
	int height=GFR_ICON_HEIGHT;
   int w2=GFR_ICON_WIDTH2;
   int h2=GFR_ICON_HEIGHT2;
	int font_width=10;//TODO:  THESE SHOULD BE CALCULATED !!!!!!
   int chars_per_line=width/(font_width>>1); //  ditto
   int len=strlen(caption);
   char *newcaption="";
	if(icon){
		if(icon->caption){
			gtk_object_destroy(GTK_OBJECT(icon->caption));
			icon->caption=NULL;
		}
		//FIXME:  about the clipping....find out about GnomeCanvasHacktext 
		//TODO:  if you don't use the above...then make this a function and be nice...*wrap* instead of _CUTTING_ words...
		if(len>chars_per_line){
			a=0;
			tmp=NULL;
			while(a<len){
				if(len-a>=chars_per_line){
					tmp=g_strndup(caption+a,chars_per_line);
					newcaption=g_strconcat(newcaption,tmp,"\n",NULL);
					g_free(tmp);
					a+=chars_per_line;
				}else{
					tmp=g_strndup(caption+a,len-a);
					newcaption=g_strconcat(newcaption,tmp,"\n",NULL);
					g_free(tmp);
					a=len;
				}
			}
			w=0;
		}else{
			newcaption=caption;
			w=((chars_per_line-len)/2)*(font_width/2);
		}
		icon->caption=gnome_canvas_item_new(GNOME_CANVAS_GROUP(icon->icon_group),gnome_canvas_text_get_type(),
			"x",(gdouble)(w-w2), "y",(gdouble)((height+2)-h2),
			"x_offset",(gdouble)0.5,"y_offset",(gdouble)0.0,
			"text",newcaption,
			"justification",GTK_JUSTIFY_CENTER,
			"fill_color_rgba",0x444444FF,
			"anchor",GTK_ANCHOR_NW,
			"font",g_strdup_printf("-*-helvetica-*-r-*-*-%i-*-*-*-*-*-*",font_width),
			//CLIPPING DOESN'T WORK RIGHT IN AA MODE....
			//		"clip",TRUE,		"clip_width",(gdouble)64.0,		"clip_height",(gdouble)32.0,
			NULL);
	}
	return(icon);
}
basic_icon* basic_icon_dup(const basic_icon*from){
	basic_icon*icon=(basic_icon*)g_memdup(from,sizeof(basic_icon));
   return(icon);
}
void basic_icon_move(basic_icon*icon,gdouble xoff,gdouble yoff){
	if(icon){
		icon->x1+=xoff;
		icon->y1+=yoff;
		icon->x2+=xoff;
		icon->y2+=yoff;
		gnome_canvas_item_move(GNOME_CANVAS_ITEM(icon->icon_group),xoff,yoff);
	}
}
void basic_icon_hide(basic_icon*icon){
   if(icon){
      gnome_canvas_item_hide(GNOME_CANVAS_ITEM(icon->icon_group));
      icon->highlighted=TRUE;
   }
}
void basic_icon_show(basic_icon*icon){
   if(icon){
      gnome_canvas_item_show(GNOME_CANVAS_ITEM(icon->icon_group));
      icon->highlighted=TRUE;
   }
}
void basic_icon_prelight(basic_icon*icon){
//   gdk_pixbuf_saturate_and_pixelate(icon->_pixbuf,icon->_pixbuf,1.25,FALSE);
   if(icon){
      gnome_canvas_item_show(GNOME_CANVAS_ITEM(icon->highlight));
		gnome_canvas_item_grab_focus(GNOME_CANVAS_ITEM(icon->icon_group));
      icon->highlighted=TRUE;
   }
}
void basic_icon_unprelight(basic_icon*icon){
//   gdk_pixbuf_saturate_and_pixelate(icon->_pixbuf,icon->_pixbuf,0.75,FALSE);
   if(icon){
      gnome_canvas_item_hide(GNOME_CANVAS_ITEM(icon->highlight));
      icon->highlighted=FALSE;
   }
}

void basic_icon_highlight(basic_icon*icon){
	if(icon){
		basic_icon_prelight(icon);
		gnome_canvas_item_set(GNOME_CANVAS_ITEM(icon->highlight),
				"fill_color_rgba",0x44BBFF22,
				NULL);
	}
}
void basic_icon_unhighlight(basic_icon*icon){
	if(icon){
		basic_icon_unprelight(icon);
		gnome_canvas_item_set(GNOME_CANVAS_ITEM(icon->highlight),
				"fill_color_rgba",0x2222FF22,
				NULL);
	}
}


void basic_icon_show_preview(basic_icon*icon){
   if(icon){
      if(icon->preview){
         gnome_canvas_item_show(GNOME_CANVAS_ITEM(icon->preview));
      }
      icon->previewing=TRUE;
   }
}
void basic_icon_hide_preview(basic_icon*icon){
   if(icon){
      if(icon->preview){
         gnome_canvas_item_hide(GNOME_CANVAS_ITEM(icon->preview));
      }
      icon->previewing=FALSE;
   }
}

//Animate a basic_icon
void basic_icon_animate(basic_icon*icon,double speed){
   double affine[6];
//TODO:  make this a run time option {
	int width=GFR_ICON_WIDTH;
	int height=GFR_ICON_HEIGHT;
// }
   int w2=width>>1;
   int h2=height>>1;
   if(!icon)return;
   affine[0]=affine[1]=affine[2]=affine[3]=affine[4]=affine[5]=0.0;
   switch(icon->effect_type){
      case ANI_BOUNCE:
         icon->effect_data[0]-=(speed*icon->effect_data[1]);
         if(icon->effect_data[0]>=0.0)icon->effect_data[1]=-icon->effect_data[1];
         if(icon->effect_data[0]<=-h2)icon->effect_data[1]=-icon->effect_data[1];
         art_affine_translate(affine,(double)0.0,(double)icon->effect_data[0]);
         gnome_canvas_item_affine_absolute(GNOME_CANVAS_ITEM(icon->pixbuf),affine);
         if(icon->link) gnome_canvas_item_affine_absolute(GNOME_CANVAS_ITEM(icon->link),affine);
         break;
      case ANI_PULSATE:
         icon->effect_data[0]+=(speed*icon->effect_data[1]);
         if(icon->effect_data[0]>1.25)icon->effect_data[1]=-icon->effect_data[1];
         if(icon->effect_data[0]<0.75)icon->effect_data[1]=-icon->effect_data[1];
         art_affine_scale(affine,(gdouble)icon->effect_data[0],(gdouble)icon->effect_data[0]);
         affine[4]=w2-w2*icon->effect_data[0];
         affine[5]=h2-h2*icon->effect_data[0];
         gnome_canvas_item_affine_absolute(GNOME_CANVAS_ITEM(icon->pixbuf),affine);
		 affine[4]=-affine[4];
		 affine[5]=-affine[5];
         if(icon->link) gnome_canvas_item_affine_absolute(GNOME_CANVAS_ITEM(icon->link),affine);
         break;
      case ANI_ROTATE:  //TODO: HACK:  ADD BREAK HERE WHEN FINISHED  (or make ANI_BOUNCE the Default)
      default:
         icon->effect_data[0]+=speed;
         if(icon->effect_data[0]>=360.0)icon->effect_data[0]-=360.0;
         art_affine_rotate(affine,(gdouble)speed);
         gnome_canvas_item_affine_relative(GNOME_CANVAS_ITEM(icon->pixbuf),affine);
         if(icon->link) gnome_canvas_item_affine_relative(GNOME_CANVAS_ITEM(icon->link),affine);
   }
}

void basic_icon_reset_animation(basic_icon*icon){
	if(icon){
		int tmp=icon->effect_type;
		icon->effect_type=-1;
		basic_icon_switch_effect(icon,tmp);//Hack to reset variables
		basic_icon_animate(icon,0.0);
	}
}

gboolean basic_icon_switch_effect(basic_icon*icon,basic_icon_AnimationMode animode){ //BREAKDOWN:  icon, desired animation mode     RETURNS:  whether effect mode was changed
   if(icon){
      if(icon->effect_type!=animode){
         switch(animode){
            case ANI_BOUNCE:
               icon->effect_data[0]=0.0;  //offset (negated for simpler math)
               icon->effect_data[1]=1.0;  //delta
               break;
            case ANI_PULSATE:
               icon->effect_data[0]=1.0;  //scale
               icon->effect_data[1]=1.0;  //delta
               break;
            default:
               icon->effect_data[0]=0.0;
               icon->effect_data[1]=0.0;
         }
         icon->effect_type=animode;
         return(TRUE);
      }
   }
   return(FALSE);
}

void basic_icon_free(basic_icon*icon){
	if(icon){
		if(icon->_pixbuf) gdk_pixbuf_unref(icon->_pixbuf);
		if(icon->_link)gdk_pixbuf_unref(icon->_link);
		if(icon->_preview_pixbuf) gdk_pixbuf_unref(icon->_preview_pixbuf);
		//TODO:  nuke objects?
		g_free((gpointer)icon);
	}
}
