//
//  GFileRunner file
//  By Kyle Davis
//
//  This controls clipboard related functions.
//

#include "gfilerunner-clipboard.h"
#include "gfilerunner-basic-icon-types.h"

gfilerunner_clipboard* gfilerunner_clipboard_new(){
	gfilerunner_clipboard*gfrclip=g_new(gfilerunner_clipboard,1);
	g_assert(gfrclip!=NULL);
	gfrclip->mode=CLIPBOARD_COPY;
	gfrclip->uri_prefix=NULL;
	gfrclip->data=NULL;
	return(gfrclip);
}

void gfilerunner_clipboard_set_prefix(gfilerunner_clipboard*gfrclip, const char* uri_prefix){
	if(gfrclip)g_free(gfrclip->uri_prefix);
	gfrclip->uri_prefix=g_strdup(uri_prefix);
}

void gfilerunner_clipboard_append_file(gfilerunner_clipboard*gfrclip,gfilerunner_vfs_obj*obj,gfilerunner_clipboard_mode mode){
	if(gfrclip){
		gfrclip->mode=mode;
		gfrclip->data=g_list_append(gfrclip->data,(gpointer)gfilerunner_vfs_obj_dup(obj));
	}
}

void gfilerunner_clipboard_prepend_file(gfilerunner_clipboard*gfrclip,gfilerunner_vfs_obj*obj,gfilerunner_clipboard_mode mode){
	if(gfrclip){
		gfrclip->mode=mode;
		gfrclip->data=g_list_prepend(gfrclip->data,(gpointer)gfilerunner_vfs_obj_dup(obj));
	}
}


gboolean gfilerunner_clipboard_has_data(gfilerunner_clipboard*gfrclip){
	if(gfrclip){
		if(gfrclip->data){
			return(TRUE);
		}
	}
	return(FALSE);
}

//FIXME:  see ToDo below
//TODO:  This will eventually need to be implimented (maybe?)
//gfilerunner_clipboard_set

void gfilerunner_clipboard_flush(gfilerunner_clipboard*gfrclip){
	GList* listitem=NULL;
	if(gfrclip){
		if(gfrclip->uri_prefix){
			g_free(gfrclip->uri_prefix);
			gfrclip->uri_prefix=NULL;
		}
		if(gfrclip->data){
			listitem=gfrclip->data;
			while(listitem){
				if(listitem->data) gfilerunner_vfs_obj_free((gfilerunner_vfs_obj*)listitem->data);
				listitem=listitem->next;
			}
			gfrclip->data=NULL;
		}
	}
}

void gfilerunner_clipboard_free(gfilerunner_clipboard*gfrclip){
	if(gfrclip){
		gfilerunner_clipboard_flush(gfrclip);
		g_free(gfrclip);
		gfrclip=NULL;
	}
}
