#include <sys/stat.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include "gfilerunner-preferences.h"

gfilerunner_preferences*gfilerunner_preferences_new(){
	char*filename;
	gfilerunner_preferences*gfrprefs=g_new(gfilerunner_preferences,1);
	gfrprefs->file=NULL;
	gfrprefs->locations=g_new(gfilerunner_locations_preferences,1);
	gfrprefs->theme=g_new(gfilerunner_theme_preferences,1);
	gfrprefs->graphics=g_new(gfilerunner_graphics_preferences,1);
	gfrprefs->view_defaults=g_new(gfilerunner_view_preferences,1);
	gfrprefs->speed=g_new(gfilerunner_speed_preferences,1);
	gfrprefs->trash=g_new(gfilerunner_trash_preferences,1);
	gfrprefs->disk_usage=g_new(gfilerunner_disk_usage_preferences,1);
	gfrprefs->userfriendly=g_new(gfilerunner_userfriendly_preferences,1);
	gfrprefs->plugin=g_new(gfilerunner_plugin_preferences,1);
	gfrprefs->ext_usage_stats=NULL;
	/* config dir creation borrowed and modified from xmms */
	filename=g_strconcat(g_get_home_dir(), "/.gfilerunner", NULL);
	mkdir(filename, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	g_free(filename);
	filename=g_strconcat(g_get_home_dir(), "/.gfilerunner/themes", NULL);
	mkdir(filename, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	g_free(filename);
	filename=g_strconcat(g_get_home_dir(), "/.gfilerunner/scripts", NULL);
	mkdir(filename, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	g_free(filename);
	filename=g_strconcat(g_get_home_dir(), "/.gfilerunner/templates", NULL);
	mkdir(filename, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	g_free(filename);
	filename=g_strconcat(g_get_home_dir(), "/.gfilerunner/plugins", NULL);
	mkdir(filename, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	g_free(filename);
	/* TODO:  Trash directory shouldn't be hard coded, should be part of prefs */
	filename=g_strconcat(g_get_home_dir(), "/Trash", NULL);
	mkdir(filename, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	g_free(filename);
/* START DEFAULTS */
	gfrprefs->theme->theme_prefix=GFILERUNNER_DEFAULT_THEME;
	gfrprefs->locations->terminal="gnome-terminal";
	gfrprefs->graphics->canvas_aa=FALSE;
	gfrprefs->graphics->AniMode=ANI_ROTATE;
/* END DEFAULTS */
	return(gfrprefs);
}

void gfilerunner_preferences_load(gfilerunner_preferences*gfrprefs,const char*file){ /* load from conf file */
	FILE*i;
	char*validblocks[]={"Global","Theme","Locations","Graphics",NULL}; /*a list of valid global block names*/
	char*validglobalkeys[]={"",NULL};
	char*validthemekeys[]={"theme","toolbar_style","toolbar_size",NULL};
	char*validlocationskeys[]={"home","trash","desktop","terminal",NULL};
	char*validgraphicskeys[]={"AntiAliasing","IconAniMode",NULL};
	char**validkeyhash[]={validglobalkeys,validthemekeys,validlocationskeys,validgraphicskeys,NULL};
	char*buf=g_malloc(1024);
	if(!file){
		gfrprefs->file=g_strconcat(g_get_home_dir(), "/.gfilerunner/config", NULL);
	}else{
		gfrprefs->file=g_strdup(file);
	}
	if((i=fopen(gfrprefs->file,"rb"))){
		int block=0;
		while(!feof(i)){
			gboolean newblock=FALSE;
			buf=fgets(buf,1024,i);
			if(buf){
				int a=0;
				int len=strlen(buf);
				if(buf[len-1]=='\n'){ /*shave endlines*/
					buf[len-1]=0;
					len--;
				}
				if(buf[len-1]=='\r'){ /*shave endlines MSDOS-style*/
					buf[len-1]=0;
					len--;
				}
				while(validblocks[a]){ /*attempt to see if the data is a valid block*/
					char*tmp=g_strdup_printf("[%s]",validblocks[a]); /* global section */
					if(strncasecmp(buf,tmp,strlen(tmp))==0){
						block=a;
						newblock=TRUE;
						break; /* if a valid block was found jump out of our loop */
					}
					g_free(tmp);
					a++;
				}
				if(!newblock){ /*as to avoid checking twice*/
					char*var=NULL;
					char*val=NULL;
					char*offset=strchr(buf,'=');
					if(offset){
						val=offset+1;
						offset[0]=0;
						var=buf;
						if(val){
							char**keylist=validkeyhash[block];
							a=0;
							while(keylist[a]){
								if(strcasecmp(keylist[a],var)==0){
									switch(block){
										case 3: /*graphics*/
											switch(a){
												case 0: /*AntiAliasing*/
													gfrprefs->graphics->canvas_aa=get_str_bool(val);
													break;
												case 1: /*IconAniMode*/
													gfrprefs->graphics->AniMode=atoi(val);
													break;
											}
											break;
										case 2: /*locations*/
											switch(a){
												case 0: /*home*/
													break;
												case 1: /*trash*/
													break;
												case 2: /*desktop*/
													break;
												case 3: /*terminal*/
													gfrprefs->locations->terminal=g_strdup(val);
													break;
											}
											break;
										case 1: /*theme*/
											switch(a){
												case 0: /*theme*/
													gfrprefs->theme->theme_prefix=g_strdup(val);
													break;
											}
											break;
										case 0: /*global*/
										default:
											break;
									}
									break;
								}
								a++;
							}
						}
					}
				}				
			}
		}
		fclose(i);
	}else{
		fprintf(stderr,"WARNING: Unable to read from config file \"%s\"!!\n",gfrprefs->file);
		/* TODO: if file!=DEFAULT_CONF_FILE then call this again with that... */
	}
	g_free(buf);
}

/* sync to config file (in gfrprefs struct) */
void gfilerunner_preferences_sync(gfilerunner_preferences*gfrprefs){
	FILE*o;
	char*hackcfg=g_strdup_printf("[global]\n\n[themes]\ntheme=%s\n\n[locations]\nterminal=%s\n\n[graphics]\nantialiasing=%i\n",
		gfrprefs->theme->theme_prefix,gfrprefs->locations->terminal,gfrprefs->graphics->canvas_aa);
	if((o=fopen(gfrprefs->file,"wb"))){
		fprintf(o,hackcfg);
		fclose(o);
	}
	g_free(hackcfg);
}
//gfilerunner_preferences_free /*duh*/
