/* gfilerunner-script.c  - GFileRunner script module handling - by Kyle Davis */
#include "gfilerunner-sys.h"
#include "gfilerunner-script.h"

gfilerunner_script*gfilerunner_script_new(const char*script){
	int a=0;
	char**tmp=NULL;
	gfilerunner_script*gfrscript=g_new(gfilerunner_script,1);
	g_assert(gfrscript!=NULL);
	gfrscript->file=g_strdup(script);
	tmp=script_get_info_array(script,GFILERUNNER_SCRIPT_GET_INFO);
	while(tmp[a]){
		switch(a){
			case 0: /*version*/
				gfrscript->version=g_strdup(tmp[a]);
				break;
			case 1: /*name*/
				gfrscript->name=g_strdup(tmp[a]);
				break;
			case 2: /*authors*/
				gfrscript->authors=g_strsplit(tmp[a],",",0);
				break;
			case 3: /*about*/
				gfrscript->about=g_strdup(tmp[a]);
				break;
			case 4: /*mime_types*/
				gfrscript->mime_types=g_strsplit(tmp[a],",",0);
				break;
			case 5: /*needs_terminal*/
				if((tmp[a])&&(strlen(tmp[a])>0)){
					gfrscript->needs_terminal=TRUE;
				}else{
					gfrscript->needs_terminal=FALSE;
				}
				break;
			case 6: /*multiple_files*/
				if((tmp[a])&&(strlen(tmp[a])>0)){
					gfrscript->multiple_files=TRUE;
				}else{
					gfrscript->multiple_files=FALSE;
				}
				break;
			case 7: /*local_only*/
				if((tmp[a])&&(strlen(tmp[a])>0)){
					gfrscript->local_only=TRUE;
				}else{
					gfrscript->local_only=FALSE;
				}
				break;
		}
		a++;
	}
	g_strfreev(tmp);
	return(gfrscript);
}

gfilerunner_script*gfilerunner_script_dup(gfilerunner_script*oldgfrscript){
	gfilerunner_script*gfrscript=g_new(gfilerunner_script,1);
	char*tmp=NULL;
	g_assert(gfrscript!=NULL);
	gfrscript->file=g_strdup(oldgfrscript->file);
	if(oldgfrscript->name) gfrscript->name=g_strdup(oldgfrscript->name);
	if(oldgfrscript->version) gfrscript->version=g_strdup(oldgfrscript->version);
	if(oldgfrscript->authors){
		tmp=g_strjoinv("\n",oldgfrscript->authors);
		gfrscript->authors=g_strsplit(tmp,"\n",0);
		g_free(tmp);
	}
	if(oldgfrscript->about) gfrscript->about=g_strdup(oldgfrscript->about);
	if(oldgfrscript->mime_types){
		tmp=g_strjoinv("\n",oldgfrscript->mime_types);
		gfrscript->mime_types=g_strsplit(tmp,"\n",0);
		g_free(tmp);
	}
	gfrscript->needs_terminal=oldgfrscript->needs_terminal;
	gfrscript->multiple_files=oldgfrscript->multiple_files;
	gfrscript->local_only=oldgfrscript->local_only;
	return(gfrscript);
}

void gfilerunner_script_free(gfilerunner_script*gfrscript){
	if(gfrscript){
		if(gfrscript->file) g_free(gfrscript->file);
		if(gfrscript->name) g_free(gfrscript->name);
		if(gfrscript->version) g_free(gfrscript->version);
		if(gfrscript->authors) g_strfreev(gfrscript->authors);
		if(gfrscript->about) g_free(gfrscript->about);
		if(gfrscript->mime_types) g_strfreev(gfrscript->mime_types);
		g_free(gfrscript);
	}
}
