/* gfilerunner-template.c  - GFileRunner template module handling - by Kyle Davis */
#include <glib.h>
#include "gfilerunner-sys.h"
#include "gfilerunner-template.h"

gfilerunner_template*gfilerunner_template_new(const char*script){
	int a=0;
	char**tmp=NULL;
	gfilerunner_template*gfrtemplate=g_new(gfilerunner_template,1);
	g_assert(gfrtemplate!=NULL);
	gfrtemplate->file=g_strdup(script);
	tmp=script_get_info_array(script,GFILERUNNER_TEMPLATE_GET_INFO);
	while(tmp[a]){
		switch(a){  /* handle which line is which */
			case 0: /*version*/
				gfrtemplate->version=g_strdup(tmp[a]);
				break;
			case 1: /*name*/
				gfrtemplate->name=g_strdup(tmp[a]);
				break;
			case 2: /*authors*/
				gfrtemplate->authors=g_strsplit(tmp[a],",",0);
				break;
			case 3: /*about*/
				gfrtemplate->about=g_strdup(tmp[a]);
				break;
			case 4: /*local_only*/
				if((tmp[a])&&(strlen(tmp[a])>0)){
					gfrtemplate->local_only=TRUE;
				}else{
					gfrtemplate->local_only=FALSE;
				}
				break;
			case 5: /*mime_type*/
				gfrtemplate->mime_type=g_strdup(tmp[a]);
				break;
			case 6: /*dependancies*/
				gfrtemplate->dependancies=g_strsplit(tmp[a],",",0);
				break;
			case 7: /*default desired name*/
				gfrtemplate->desired=g_strdup(tmp[a]);
				break;
		}
		a++;
	}
	g_strfreev(tmp);
	return(gfrtemplate);
}

gfilerunner_template*gfilerunner_template_dup(gfilerunner_template*oldgfrtemplate){
	gfilerunner_template*gfrtemplate=g_new(gfilerunner_template,1);
	char*tmp=NULL;
	g_assert(gfrtemplate!=NULL);
	gfrtemplate->file=g_strdup(oldgfrtemplate->file);
	if(oldgfrtemplate->name) gfrtemplate->name=g_strdup(oldgfrtemplate->name);
	if(oldgfrtemplate->version) gfrtemplate->version=g_strdup(oldgfrtemplate->version);
	if(oldgfrtemplate->authors){
		tmp=g_strjoinv(",",oldgfrtemplate->authors);
		gfrtemplate->authors=g_strsplit(tmp,",",0);
		g_free(tmp);
	}
	if(oldgfrtemplate->about) gfrtemplate->about=g_strdup(oldgfrtemplate->about);
	if(oldgfrtemplate->dependancies){
		tmp=g_strjoinv(",",oldgfrtemplate->dependancies);
		gfrtemplate->dependancies=g_strsplit(tmp,",",0);
		g_free(tmp);
	}
	if(gfrtemplate->mime_type) gfrtemplate->mime_type=g_strdup(oldgfrtemplate->mime_type);
	if(oldgfrtemplate->desired) gfrtemplate->desired=g_strdup(oldgfrtemplate->desired);
	gfrtemplate->local_only=oldgfrtemplate->local_only;
	return(gfrtemplate);
}

char**gfilerunner_template_check_dependancies(gfilerunner_template*gfrtemplate){
	/* if good return NULL, otherwise return where it failed */
	char**depn=NULL;
	char**deplist=NULL;
	int depn_offset=0;
	int deplist_offset;
	char*ret=NULL;
	char*depslist=NULL; /* comma seperated list of deps to use */
	if(gfrtemplate->dependancies){
		depn=gfrtemplate->dependancies;
		while(depn[depn_offset]){
			deplist=g_strsplit(depn[depn_offset],"|",0);
			deplist_offset=0;
			ret=NULL;
			while((deplist[deplist_offset])&&(!ret)){
				/* a slight hack, could use gnomevfs, but don't want the dependancy here, this should work for now */
				ret=script_get_info_var("which",deplist[deplist_offset]);
				deplist_offset++;
			}
			if(!ret)ret=g_strdup(GFILERUNNER_TEMPLATE_ERROR);
			if(ret[strlen(ret)-1]=='\n')ret[strlen(ret)-1]=0;
			if(depslist){
				depslist=g_strconcat(depslist,",",g_strdup(ret),NULL);
			}else{
				depslist=g_strdup(ret);
			}
			g_free(ret);
			depn_offset++;
		}
	}
	return(g_strsplit(depslist,",",0));
}

void gfilerunner_template_free(gfilerunner_template*gfrtemplate){
	if(gfrtemplate){
		if(gfrtemplate->file) g_free(gfrtemplate->file);
		if(gfrtemplate->name) g_free(gfrtemplate->name);
		if(gfrtemplate->version) g_free(gfrtemplate->version);
		if(gfrtemplate->authors) g_strfreev(gfrtemplate->authors);
		if(gfrtemplate->about) g_free(gfrtemplate->about);
		if(gfrtemplate->dependancies) g_free(gfrtemplate->dependancies);
		if(gfrtemplate->mime_type) g_free(gfrtemplate->mime_type);
		if(gfrtemplate->desired) g_free(gfrtemplate->desired);
		g_free(gfrtemplate);
	}
}
