// gfilerunner-tree.c  -   by Kyle Davis
//
// This should allow GFileRunner and its modules handle file/directory trees properly.
//
#include <glib.h>
#include "gfilerunner-tree.h"

gfilerunner_tree* gfilerunner_tree_new(GtkWidget*ctree,const char*collapse_pixbuf_file,const char*expand_pixbuf_file){
	gfilerunner_tree*gfrtree=g_new(gfilerunner_tree,1);
	gfilerunner_tree_object*gfrtree_obj;
	g_assert(gfrtree!=NULL);
	g_assert(ctree!=NULL);
	gfrtree->uri_desc=NULL;
	gfrtree_obj=gfilerunner_tree_object_new("ROOT","ROOT",collapse_pixbuf_file,expand_pixbuf_file);
	gfrtree->rootnode=g_node_new((gpointer)gfrtree_obj);
	gfrtree->ctree=ctree;
	return(gfrtree);	
}

void gfilerunner_tree_register_uri_desc(gfilerunner_tree*gfrtree,gfilerunner_uri_desc*uridesc){
	g_assert(gfrtree!=NULL);
	gfrtree->uri_desc=g_list_append(gfrtree->uri_desc,(gpointer)uridesc);
}

gboolean uri_has_parent(const char*uri){
	GnomeVFSURI* curi=gnome_vfs_uri_new(uri);
	gboolean ret=gnome_vfs_uri_has_parent(curi);
	gnome_vfs_uri_unref(curi);
	return(ret);
}

char*get_uri_parent(const char*uri){
	//use GNOME VFS to do this for us :)
	GnomeVFSURI* curi=gnome_vfs_uri_new(uri);
	GnomeVFSURI* puri=NULL;
	char*puri_txt=NULL;
	if(uri_has_parent(uri)){
		puri=gnome_vfs_uri_get_parent(curi);
		puri_txt=gnome_vfs_uri_to_string(puri,GNOME_VFS_URI_HIDE_USER_NAME&GNOME_VFS_URI_HIDE_PASSWORD&
							GNOME_VFS_URI_HIDE_HOST_NAME&GNOME_VFS_URI_HIDE_HOST_PORT);
		gnome_vfs_uri_unref(puri);
		gnome_vfs_uri_unref(curi);
		return(puri_txt);
	}else{
		gnome_vfs_uri_unref(curi);
	}
	return NULL;
}

GList*get_uri_parents_reversed(gfilerunner_uri_desc*uri_desc,const char* location){
	GList*plist=NULL;
	char*parent=uri_desc->modifier((gpointer)uri_desc,location);
	while(parent!=NULL){
//len=strlen(parent);
//if(parent[len-1]!=GNOME_VFS_URI_PATH_CHR) parent=g_strconcat(parent,GNOME_VFS_URI_PATH_STR,NULL);
		plist=g_list_prepend(plist,(gpointer)uri_desc->modifier((gpointer)uri_desc,parent));
		parent=get_uri_parent(parent);
	}
	return(plist);
}

GNode* gfilerunner_tree_find_node(GNode*node,const char*location){
	GNode*child=NULL;
	GNode*recurse=NULL;
	gfilerunner_tree_object*gfrtree_object=NULL;
	g_assert(node!=NULL);
	g_assert(node->data!=NULL);
	child=node->children;
	while(child!=NULL){
		gfrtree_object=(gfilerunner_tree_object*)child->data;
		if((gfrtree_object!=NULL)&&(gfrtree_object->location!=NULL)){
			if(strcmp(gfrtree_object->location,location)==0){
				return(child);
			}else{
				recurse=gfilerunner_tree_find_node(child,location);
				if(recurse) return(recurse);
			}
		}
		child=child->next;
	}
	return(NULL);
}

gfilerunner_tree_object* gfilerunner_tree_register_folder(gfilerunner_tree*gfrtree,const char* location){
	GNode*myparent=NULL;
	GNode*findret=NULL;
	GList*myparentlist=NULL;
	GList*myuridesclst=NULL;
	gfilerunner_uri_desc*myuridesc=NULL;
	GtkCTreeNode*ctree_parentnode=NULL;
	GtkCTreeNode*ctree_newnode=NULL;
	gfilerunner_tree_object*gfrtree_obj=NULL; //a temp object really...this will hopefully reduce some confusion...
	gfilerunner_tree_object*rootobj=(gfilerunner_tree_object*)gfrtree->rootnode->data;
	char*mylocation=NULL;
	char*text[]={NULL,NULL,NULL};
	g_assert(gfrtree!=NULL);
	g_assert(rootobj!=NULL);
	myparent=gfrtree->rootnode;
//len=strlen(mylocation);
//if(mylocation[len-1]!=GNOME_VFS_URI_PATH_CHR) mylocation=g_strconcat(mylocation,GNOME_VFS_URI_PATH_STR,NULL);

//URI Type Detection
	myuridesclst=gfrtree->uri_desc;
	while(myuridesclst){
		myuridesc=(gfilerunner_uri_desc*)myuridesclst->data;
		if(strncasecmp(myuridesc->prefix,location,strlen(myuridesc->prefix))==0) break;	//we found a match...we're done here
		myuridesclst=myuridesclst->next;
	}
	mylocation=myuridesc->modifier((gpointer)myuridesc,location);

//TODO:  Soon you should do Sorting...
//TODO:  Figure out how to give each object a copy of the root object's GdkBitmap & GdkPixmap
	myparentlist=get_uri_parents_reversed(myuridesc,mylocation);
	while(myparentlist){
		findret=gfilerunner_tree_find_node(myparent,(char*)myparentlist->data);
		if(!findret){
			text[0]=myuridesc->humanize((gpointer)myuridesc,(char*)myparentlist->data);
			gfrtree_obj=gfilerunner_tree_object_new_noload((char*)myparentlist->data,text[0],
					rootobj->collapse_pixmap,rootobj->collapse_mask,rootobj->expand_pixmap,rootobj->expand_mask);
			if(!ctree_parentnode){
				ctree_newnode=gtk_ctree_insert_node(GTK_CTREE(gfrtree->ctree),NULL,NULL,text,2,
					rootobj->collapse_pixmap,rootobj->collapse_mask,rootobj->expand_pixmap,rootobj->expand_mask,FALSE,TRUE);
			}else{
				ctree_newnode=gtk_ctree_insert_node(GTK_CTREE(gfrtree->ctree),GTK_CTREE_NODE(ctree_parentnode),NULL,text,2,
					rootobj->collapse_pixmap,rootobj->collapse_mask,rootobj->expand_pixmap,rootobj->expand_mask,FALSE,FALSE);
			}
			gfrtree_obj->data=(gpointer)ctree_newnode;
			gtk_ctree_node_set_row_data(GTK_CTREE(gfrtree->ctree),GTK_CTREE_NODE(ctree_newnode),(gpointer)gfrtree_obj);
			myparent=g_node_insert_data(myparent,-1,(gpointer)gfrtree_obj);
		}else{
			myparent=findret;
		}
		gfrtree_obj=(gfilerunner_tree_object*)myparent->data;
		ctree_parentnode=(GtkCTreeNode*)gfrtree_obj->data;
		g_free(myparentlist->data);	//We're done with you now...go...be free!
		myparentlist=myparentlist->next;
	}
	g_free(mylocation);
	g_list_free(myparentlist);
	return(gfrtree_obj);
}

void gfilerunner_tree_unregister_folder(gfilerunner_tree*gfrtree,const char* location){
	//find location in (gfrtree_object*)(gfrtree->rootnode)->....remove it and all children
	//sync with ctree
}

//clean up....
int gfilerunner_tree_register_special(gfilerunner_tree*gfrtree,const char* vlocation,char* desc,GdkPixbuf*pixbuf,char*text){
	return(-1);//not_done
}

void gfilerunner_tree_free(gfilerunner_tree*gfrtree){
	g_assert(gfrtree!=NULL);
	//free
}
