// gfilerunner-tree.h  -   by Kyle Davis
//
// This should allow GFileRunner and its modules handle file/directory trees properly.
//
#ifndef GFILERUNNER_TREE_H
#define GFILERUNNER_TREE_H

#include <libgnomevfs/gnome-vfs.h>
#include "gfilerunner-tree-object.h"
#include "gfilerunner-uri.h"

typedef struct _gfilerunner_tree gfilerunner_tree;
struct _gfilerunner_tree{
	GList*uri_desc; //GList of uri descriptions ie (file:/// == My Computer)
	GNode*rootnode;			//GNode of gfilerunner_tree_object's
	GtkWidget*ctree;		//Real ctree widget
};

//New gfilerunner_tree
gfilerunner_tree* gfilerunner_tree_new(GtkWidget*ctree,const char*collapse_pixbuf_file,const char*expand_pixbuf_file);

//Connect a URI description to a gfilerunner_tree
void gfilerunner_tree_register_uri_desc(gfilerunner_tree*gfrtree,gfilerunner_uri_desc*uridesc);

//Helps to find a node by the location
gboolean uri_has_parent(const char*uri);
char*get_uri_parent(const char*uri);
GList*get_uri_parents_reversed(gfilerunner_uri_desc*uri_desc,const char* location);
GNode* gfilerunner_tree_find_node(GNode*node,const char*location);

//Registers a folder in the tree.
//Position is determined by # of seperators (above) in the location
//Text is the last section of text after last seperator, if that is empty then the next found will be used.
gfilerunner_tree_object* gfilerunner_tree_register_folder(gfilerunner_tree*gfrtree,const char* location);

//Unregister a folder in the tree and all of its sub items
void gfilerunner_tree_unregister_folder(gfilerunner_tree*gfrtree,const char* location);


//TODO:  Add special pixmaps for special folders, such as:   Home, Trash, Documents
//Register something that doesn't really exist.
//Will be placed according to the vlocation
//Events must be handled by code specifically.
int gfilerunner_tree_register_special(gfilerunner_tree*gfrtree,const char* vlocation,char* desc,GdkPixbuf*pixbuf,char*text);

//Free all resources used by tree
void gfilerunner_tree_free(gfilerunner_tree*gfrtree);

#endif
